import inspect
import os
import re
from typing import List

import pydantic
from langchain.agents.load_tools import (
    _BASE_TOOLS,
    _EXTRA_LLM_TOOLS,
    _EXTRA_OPTIONAL_TOOLS,
    _LLM_TOOLS,
)
from langchain.llms import OpenAI
from pydantic import BaseModel, Field

_all_tools = {**_BASE_TOOLS, **_EXTRA_LLM_TOOLS, **_EXTRA_OPTIONAL_TOOLS, **_LLM_TOOLS}


def get_dummy_token():
    return 'dummy_token'


def get_dummy_llm():
    os.environ['OPENAI_API_KEY'] = get_dummy_token()
    return OpenAI(temperature=0)


def missing_key_from_err(error):
    validation_err_pattern = r"\`([a-zA-Z0-9_]+)\`"
    matches = re.findall(validation_err_pattern, error)
    return [m for m in matches if all(c.islower() or c == '_' for c in m)]


class LangchainTool(BaseModel):
    name: str
    api: str
    args: List[str] = Field(default_factory=list)


def get_all_langchain_tools() -> List[LangchainTool]:
    l_tools = []
    for k, v in _all_tools.items():
        args = []
        if isinstance(v, tuple):
            func = v[0]
            keys = v[1:][0]
        else:
            func = v
            keys = []

        sig = inspect.signature(func)
        if len(sig.parameters) == 0:
            name = func().name
        elif 'llm' in sig.parameters:
            try:
                name = func(get_dummy_llm()).name
            except KeyError as e:
                missing_key = e.args[0]
                name = func(get_dummy_llm(), **{missing_key: get_dummy_token()}).name
                args = [missing_key]
        elif 'kwargs' in sig.parameters:
            try:
                name = func().name
            except pydantic.ValidationError as e:
                keys = missing_key_from_err(e.errors()[0]['msg'])
                try:
                    name = func(**{k: get_dummy_token() for k in keys}).name
                    args = keys
                except pydantic.ValidationError as e:
                    more_keys = missing_key_from_err(e.errors()[0]['msg'])
                    name = func(**{k: get_dummy_token() for k in keys + more_keys}).name
                    args = keys + more_keys

        l_tools.append(LangchainTool(name=name, api=k, args=args))

    return l_tools


ALL_TOOLS = {t.name: {'api': t.api, 'args': t.args} for t in get_all_langchain_tools()}
ALL_TOOL_NAMES = [t.name for t in get_all_langchain_tools()]

# https://python.langchain.com/en/latest/modules/agents/agents.html
ALL_AGENT_TYPES = {
    'MRKL': 'zero-shot-react-description',
    'ReAct': 'react-docstore',
    'Self-Ask-With-Search': 'self-ask-with-search',
}



def _ignore_warnings():
    import logging
    import warnings

    logging.captureWarnings(True)
    warnings.filterwarnings(
        "ignore",
        category=DeprecationWarning,
        message="Deprecated call to `pkg_resources.declare_namespace('google')`.",
    )


_ignore_warnings()

from .backend import serving, slackbot, download_df, upload_df
from .backend.slackbot.memory import MemoryMode, get_memory

__version__ = '0.0.48'



from typing import Any, Dict, List

from llama_index.langchain_helpers.memory_wrapper import GPTIndexChatMemory
from langchain.schema import ChatMessage


class GPTMultiUserChatMemory(GPTIndexChatMemory):
    @staticmethod
    def msg_to_txt(msg: ChatMessage) -> str:
        return f'{msg.role}: {msg.content}'

    def save_context(self, inputs: Dict[str, Any], outputs: Dict[str, str]) -> None:
        from llama_index import Document as LlamaDocument
        from llama_index.utils import get_new_id

        first_role = list(inputs.keys())[0]
        first_content = list(inputs.values())[0]
        first_msg = ChatMessage(role=first_role, content=first_content)
        first_message_id = get_new_id(set(self.id_to_message.keys()))
        self.chat_memory.messages.append(first_msg)
        self.id_to_message[first_message_id] = first_msg
        first_doc = LlamaDocument(
            text=GPTMultiUserChatMemory.msg_to_txt(first_msg),
            doc_id=first_message_id,
        )

        second_role = list(outputs.keys())[0]
        second_content = list(outputs.values())[0]
        second_msg = ChatMessage(role=second_role, content=second_content)
        second_message_id = get_new_id(set(self.id_to_message.keys()))
        self.chat_memory.messages.append(second_msg)
        self.id_to_message[second_message_id] = second_msg
        second_doc = LlamaDocument(
            text=GPTMultiUserChatMemory.msg_to_txt(second_msg),
            doc_id=second_message_id,
        )

        self.index.insert(first_doc)
        self.index.insert(second_doc)



import os
import sys
from typing import List

import click
from jcloud.constants import Phase
from jina import Flow

from . import __version__
from .config import resolve_jcloud_config, validate_jcloud_config_callback
from .flow import (
    APP_NAME,
    AUTOGPT_APP_NAME,
    BABYAGI_APP_NAME,
    DEFAULT_TIMEOUT,
    PANDAS_AI_APP_NAME,
    PDF_QNA_APP_NAME,
    SLACK_BOT_NAME,
    SLACKBOT_DEMO_APP_NAME,
    ExportKind,
    create_slack_app_manifest,
    deploy_app_on_jcloud,
    get_app_status_on_jcloud,
    get_flow_dict,
    get_flow_yaml,
    get_module_dir,
    get_uri,
    list_apps_on_jcloud,
    load_local_df,
    remove_app_on_jcloud,
    syncify,
    update_requirements,
)


def serve_locally(
    module_str: str = None,
    fastapi_app_str: str = None,
    port: int = 8080,
    env: str = None,
):
    sys.path.append(os.getcwd())
    f_yaml = get_flow_yaml(
        module_str=module_str,
        fastapi_app_str=fastapi_app_str,
        jcloud=False,
        port=port,
        env=env,
    )
    with Flow.load_config(f_yaml) as f:
        # TODO: add local description
        f.block()


async def serve_on_jcloud(
    module_str: str = None,
    fastapi_app_str: str = None,
    app_dir: str = None,
    uses: str = None,
    name: str = APP_NAME,
    requirements: List[str] = None,
    app_id: str = None,
    version: str = 'latest',
    timeout: int = DEFAULT_TIMEOUT,
    platform: str = None,
    config: str = None,
    cors: bool = True,
    env: str = None,
    verbose: bool = False,
    public: bool = False,
    lcserve_app: bool = False,
) -> str:
    from .backend.playground.utils.helper import get_random_tag
    from .flow import push_app_to_hubble

    module_dir, is_websocket = get_module_dir(
        module_str=module_str,
        fastapi_app_str=fastapi_app_str,
        app_dir=app_dir,
        lcserve_app=lcserve_app,
    )
    config = resolve_jcloud_config(config, module_dir)

    if uses is not None:
        gateway_id = uses
    else:
        gateway_id = push_app_to_hubble(
            module_dir=module_dir,
            requirements=requirements,
            tag=get_random_tag(),
            version=version,
            platform=platform,
            verbose=verbose,
            public=public,
        )

    app_id, _ = await deploy_app_on_jcloud(
        flow_dict=get_flow_dict(
            module_str=module_str,
            fastapi_app_str=fastapi_app_str,
            jcloud=True,
            port=8080,
            name=name,
            timeout=timeout,
            app_id=app_id,
            gateway_id=gateway_id,
            is_websocket=is_websocket,
            jcloud_config_path=config,
            cors=cors,
            env=env,
            lcserve_app=lcserve_app,
        ),
        app_id=app_id,
        verbose=verbose,
    )
    await get_app_status_on_jcloud(app_id=app_id)
    return app_id


async def serve_babyagi_on_jcloud(
    name: str = BABYAGI_APP_NAME,
    requirements: List[str] = None,
    app_id: str = None,
    version: str = 'latest',
    timeout: int = DEFAULT_TIMEOUT,
    platform: str = None,
    config: str = None,
    cors: bool = True,
    env: str = None,
    verbose: bool = False,
    public: bool = False,
):
    requirements = requirements or []
    update_requirements(
        path=os.path.join(
            os.path.dirname(__file__), 'apps', 'babyagi', 'requirements.txt'
        ),
        requirements=requirements,
    )
    await serve_on_jcloud(
        module_str='lcserve.apps.babyagi.app',
        name=name,
        requirements=requirements,
        app_id=app_id,
        app_dir=os.path.join(os.path.dirname(__file__), 'apps', 'babyagi'),
        version=version,
        timeout=timeout,
        platform=platform,
        config=config,
        cors=cors,
        env=env,
        verbose=verbose,
        public=public,
        lcserve_app=True,
    )


async def serve_autogpt_on_jcloud(
    name: str = AUTOGPT_APP_NAME,
    requirements: List[str] = None,
    app_id: str = None,
    version: str = 'latest',
    timeout: int = DEFAULT_TIMEOUT,
    platform: str = None,
    config: str = None,
    cors: bool = True,
    env: str = None,
    verbose: bool = False,
    public: bool = False,
):
    requirements = requirements or []
    update_requirements(
        path=os.path.join(
            os.path.dirname(__file__), 'apps', 'autogpt', 'requirements.txt'
        ),
        requirements=requirements,
    )
    await serve_on_jcloud(
        module_str='lcserve.apps.autogpt.app',
        name=name,
        requirements=tuple(requirements),
        app_id=app_id,
        app_dir=os.path.join(os.path.dirname(__file__), 'apps', 'autogpt'),
        version=version,
        timeout=timeout,
        platform=platform,
        config=config,
        env=env,
        cors=cors,
        verbose=verbose,
        public=public,
        lcserve_app=True,
    )


async def serve_pdf_qna_on_jcloud(
    name: str = PDF_QNA_APP_NAME,
    app_id: str = None,
    requirements: List[str] = None,
    version: str = 'latest',
    timeout: int = DEFAULT_TIMEOUT,
    platform: str = None,
    config: str = None,
    cors: bool = True,
    env: str = None,
    verbose: bool = False,
    public: bool = False,
):
    await serve_on_jcloud(
        module_str='lcserve.apps.pdf_qna.app',
        name=name,
        app_id=app_id,
        app_dir=os.path.join(os.path.dirname(__file__), 'apps', 'pdf_qna'),
        requirements=requirements,
        version=version,
        timeout=timeout,
        platform=platform,
        config=config,
        cors=cors,
        env=env,
        verbose=verbose,
        public=public,
        lcserve_app=True,
    )


async def serve_pandas_ai_on_jcloud(
    name: str = PANDAS_AI_APP_NAME,
    app_id: str = None,
    requirements: List[str] = None,
    version: str = 'latest',
    timeout: int = DEFAULT_TIMEOUT,
    platform: str = None,
    config: str = None,
    cors: bool = True,
    env: str = None,
    verbose: bool = False,
    public: bool = False,
):
    await serve_on_jcloud(
        module_str='lcserve.apps.pandas_ai.api',
        name=name,
        app_id=app_id,
        app_dir=os.path.join(os.path.dirname(__file__), 'apps', 'pandas_ai'),
        requirements=requirements,
        version=version,
        timeout=timeout,
        platform=platform,
        config=config,
        cors=cors,
        env=env,
        verbose=verbose,
        public=public,
        lcserve_app=True,
    )


async def serve_slackbot_demo_on_jcloud(
    name: str = SLACKBOT_DEMO_APP_NAME,
    app_id: str = None,
    requirements: List[str] = None,
    version: str = 'latest',
    timeout: int = DEFAULT_TIMEOUT,
    platform: str = None,
    config: str = None,
    cors: bool = True,
    env: str = None,
    verbose: bool = False,
    public: bool = False,
):
    await serve_on_jcloud(
        module_str='lcserve.apps.slackbot.app',
        name=name,
        app_id=app_id,
        app_dir=os.path.join(os.path.dirname(__file__), 'apps', 'slackbot'),
        requirements=requirements,
        version=version,
        timeout=timeout,
        platform=platform,
        config=config,
        cors=cors,
        env=env,
        verbose=verbose,
        public=public,
        lcserve_app=True,
    )


def upload_df_to_jcloud(module: str, name: str):
    from . import upload_df

    df = load_local_df(module)
    df_id = upload_df(df, name)
    click.echo(
        "Uploaded dataframe with ID " + click.style(df_id, fg="green", bold=True)
    )


_hubble_push_options = [
    click.option(
        '--image-name',
        type=str,
        required=False,
        help='Name of the image to be pushed.',
    ),
    click.option(
        '--image-tag',
        type=str,
        default='latest',
        required=False,
        help='Tag of the image to be pushed.',
    ),
    click.option(
        '--platform',
        type=str,
        required=False,
        help='Platform of Docker image needed for the deployment is built on.',
    ),
    click.option(
        '--requirements',
        default=None,
        type=str,
        help='''Pass either

            1) multiple requirements or,
            2) a path to a requirements.txt/pyproject.toml file or,
            3) a directory containing requirements.txt/pyproject.toml file.''',
        multiple=True,
    ),
    click.option(
        '--version',
        type=str,
        default='latest',
        help='Version of serving gateway to be used.',
        show_default=False,
    ),
    click.option(
        '--verbose',
        is_flag=True,
        help='Verbose mode.',
        show_default=True,
    ),
    click.option(
        '--public',
        is_flag=True,
        help='Push the image publicly.',
        default=False,
        show_default=True,
    ),
]


_jcloud_shared_options = [
    click.option(
        '--app-id',
        type=str,
        default=None,
        help='AppID of the deployed agent to be updated.',
        show_default=True,
    ),
    click.option(
        '--requirements',
        default=None,
        type=str,
        help='''Pass either 
        1) multiple requirements or,
        2) a path to a requirements.txt/pyproject.toml file or,
        3) a directory containing requirements.txt/pyproject.toml file.''',
        multiple=True,
    ),
    click.option(
        '--version',
        type=str,
        default='latest',
        help='Version of serving gateway to be used.',
        show_default=False,
    ),
    click.option(
        '--timeout',
        type=int,
        default=DEFAULT_TIMEOUT,
        help='Total request timeout in seconds.',
        show_default=True,
    ),
    click.option(
        '--platform',
        type=str,
        default=None,
        help='Platform of Docker image needed for the deployment is built on.',
        show_default=False,
    ),
    click.option(
        '--config',
        type=click.Path(exists=True),
        help='Path to the config file',
        callback=validate_jcloud_config_callback,
        show_default=False,
    ),
    click.option(
        '--env',
        type=click.Path(exists=True),
        help='Path to the environment file',
        show_default=False,
    ),
    click.option(
        '--cors',
        is_flag=True,
        help='Enable CORS.',
        default=True,
        show_default=True,
    ),
    click.option(
        '--verbose',
        is_flag=True,
        help='Verbose mode.',
        show_default=True,
    ),
    click.option(
        '--public',
        is_flag=True,
        help='Push the image publicly.',
        default=False,
        show_default=True,
    ),
]


def hubble_push_options(func):
    for option in reversed(_hubble_push_options):
        func = option(func)
    return func


def jcloud_shared_options(func):
    for option in reversed(_jcloud_shared_options):
        func = option(func)
    return func


@click.group()
@click.version_option(__version__, '-v', '--version', prog_name='lc-serve')
@click.help_option('-h', '--help')
def serve():
    pass


@serve.command(help='Push the app image to Jina AI Cloud.')
@click.argument(
    'module_str',
    type=str,
    required=False,
)
@click.option(
    '--app',
    type=str,
    required=False,
    help='FastAPI application to run, in the format "<module>:<attribute>"',
)
@click.option(
    '--app-dir',
    type=str,
    required=False,
    help='Base directory to be used for the FastAPI app.',
)
@hubble_push_options
@click.help_option('-h', '--help')
def push(
    module_str,
    app,
    app_dir,
    image_name,
    image_tag,
    platform,
    requirements,
    version,
    verbose,
    public,
):
    from .flow import push_app_to_hubble

    module_dir, _ = get_module_dir(
        module_str=module_str,
        fastapi_app_str=app,
        app_dir=app_dir,
    )
    gateway_id = push_app_to_hubble(
        module_dir=module_dir,
        image_name=image_name,
        tag=image_tag,
        platform=platform,
        requirements=requirements,
        version=version,
        verbose=verbose,
        public=public,
    )
    id, tag = gateway_id.split(':')
    click.echo(
        f'Pushed to Hubble. Use {click.style(get_uri(id, tag), fg="green")} to deploy.'
    )


@serve.command(help='Export the app for self-hosted deployment.')
@click.argument(
    'module_str',
    type=str,
    required=False,
)
@click.option(
    '--app',
    type=str,
    required=False,
    help='FastAPI application to run, in the format "<module>:<attribute>"',
)
@click.option(
    '--app-dir',
    type=str,
    required=False,
    help='Base directory to be used for the FastAPI app.',
)
@click.option(
    '--kind',
    type=click.Choice([e.value for e in ExportKind]),
    default=ExportKind.KUBERNETES.value,
    help='Export to Kubernetes or Docker Compose.',
    show_default=True,
)
@click.option(
    '--path',
    type=str,
    default='.',
    help='Path to the directory where the export should be saved.',
    show_default=True,
)
@click.option(
    '--name',
    type=str,
    default=APP_NAME,
    help='Name of the app.',
    show_default=True,
)
@click.option(
    '--timeout',
    type=int,
    default=DEFAULT_TIMEOUT,
    help='Total request timeout in seconds.',
    show_default=True,
)
@click.option(
    '--env',
    type=click.Path(exists=True),
    help='Path to the environment file',
    show_default=False,
)
@hubble_push_options
@click.help_option('-h', '--help')
def export(
    module_str,
    app,
    app_dir,
    kind,
    path,
    name,
    timeout,
    env,
    image_name,
    image_tag,
    platform,
    requirements,
    version,
    verbose,
    public,
):
    from .flow import export_app

    export_app(
        module_str=module_str,
        fastapi_app_str=app,
        app_dir=app_dir,
        kind=kind,
        path=path,
        image_name=image_name,
        tag=image_tag,
        platform=platform,
        requirements=requirements,
        version=version,
        verbose=verbose,
        public=public,
        name=name,
        timeout=timeout,
        env=env,
    )


@serve.group(help='Deploy the app.')
@click.help_option('-h', '--help')
def deploy():
    pass


@deploy.command(help='Deploy the app locally.')
@click.argument(
    'module_str',
    type=str,
    required=False,
)
@click.option(
    '--app',
    type=str,
    required=False,
    help='FastAPI application to run, in the format "<module>:<attribute>"',
)
@click.option(
    '--port',
    type=int,
    default=8080,
    help='Port to run the server on.',
)
@click.option(
    '--env',
    type=click.Path(exists=True),
    help='Path to the environment file',
    show_default=False,
)
@click.help_option('-h', '--help')
def local(module_str, app, port, env):
    serve_locally(module_str=module_str, fastapi_app_str=app, port=port, env=env)


@deploy.command(help='Deploy the app on JCloud.')
@click.argument(
    'module_str',
    type=str,
    required=False,
)
@click.option(
    '--app',
    type=str,
    required=False,
    help='FastAPI application to run, in the format "<module>:<attribute>"',
)
@click.option(
    '--app-dir',
    type=str,
    required=False,
    help='Base directory to be used for the FastAPI app.',
)
@click.option(
    '--uses',
    type=str,
    default=None,
    help='Pass a pre-existing image that was pushed using `push-only` option.',
)
@click.option(
    '--name',
    type=str,
    default=APP_NAME,
    help='Name of the app.',
    show_default=True,
)
@jcloud_shared_options
@click.help_option('-h', '--help')
@syncify
async def jcloud(
    module_str,
    app,
    app_dir,
    uses,
    name,
    app_id,
    requirements,
    version,
    timeout,
    platform,
    config,
    cors,
    env,
    verbose,
    public,
):
    await serve_on_jcloud(
        module_str=module_str,
        fastapi_app_str=app,
        app_dir=app_dir,
        uses=uses,
        name=name,
        app_id=app_id,
        requirements=requirements,
        version=version,
        timeout=timeout,
        platform=platform,
        config=config,
        env=env,
        cors=cors,
        verbose=verbose,
        public=public,
    )


@deploy.command(help='Deploy babyagi on JCloud.')
@click.option(
    '--name',
    type=str,
    default=BABYAGI_APP_NAME,
    help='Name of the app.',
    show_default=True,
)
@jcloud_shared_options
@click.help_option('-h', '--help')
@syncify
async def babyagi(
    name,
    app_id,
    requirements,
    version,
    timeout,
    platform,
    config,
    cors,
    env,
    verbose,
    public,
):
    await serve_babyagi_on_jcloud(
        name=name,
        requirements=requirements,
        app_id=app_id,
        version=version,
        timeout=timeout,
        platform=platform,
        config=config,
        cors=cors,
        env=env,
        verbose=verbose,
        public=public,
    )


@deploy.command(help='Deploy pdf qna on JCloud.')
@click.option(
    '--name',
    type=str,
    default=PDF_QNA_APP_NAME,
    help='Name of the app.',
    show_default=True,
)
@jcloud_shared_options
@click.help_option('-h', '--help')
@syncify
async def pdf_qna(
    name,
    app_id,
    requirements,
    version,
    timeout,
    platform,
    config,
    cors,
    env,
    verbose,
    public,
):
    await serve_pdf_qna_on_jcloud(
        name=name,
        app_id=app_id,
        requirements=requirements,
        version=version,
        timeout=timeout,
        config=config,
        platform=platform,
        cors=cors,
        env=env,
        verbose=verbose,
        public=public,
    )


@deploy.command(help='Deploy autogpt on JCloud.')
@click.option(
    '--name',
    type=str,
    default=AUTOGPT_APP_NAME,
    help='Name of the app.',
    show_default=True,
)
@jcloud_shared_options
@click.help_option('-h', '--help')
@syncify
async def autogpt(
    name,
    app_id,
    requirements,
    version,
    timeout,
    platform,
    config,
    cors,
    env,
    verbose,
    public,
):
    await serve_autogpt_on_jcloud(
        name=name,
        requirements=requirements,
        app_id=app_id,
        version=version,
        timeout=timeout,
        platform=platform,
        config=config,
        cors=cors,
        env=env,
        verbose=verbose,
        public=public,
    )


@deploy.command(help='Deploy pandas-ai on JCloud.')
@click.option(
    '--name',
    type=str,
    default=PANDAS_AI_APP_NAME,
    help='Name of the app.',
    show_default=True,
)
@jcloud_shared_options
@click.help_option('-h', '--help')
@syncify
async def pandas_ai(
    name,
    app_id,
    requirements,
    version,
    timeout,
    platform,
    config,
    cors,
    env,
    verbose,
    public,
):
    await serve_pandas_ai_on_jcloud(
        name=name,
        app_id=app_id,
        requirements=requirements,
        version=version,
        timeout=timeout,
        platform=platform,
        config=config,
        cors=cors,
        env=env,
        verbose=verbose,
        public=public,
    )


@deploy.command(help='Deploy slackbot-demo on JCloud.')
@click.option(
    '--name',
    type=str,
    default=SLACKBOT_DEMO_APP_NAME,
    help='Name of the app.',
    show_default=True,
)
@jcloud_shared_options
@click.help_option('-h', '--help')
@syncify
async def slackbot_demo(
    name,
    app_id,
    requirements,
    version,
    timeout,
    platform,
    config,
    cors,
    env,
    verbose,
    public,
):
    await serve_slackbot_demo_on_jcloud(
        name=name,
        app_id=app_id,
        requirements=requirements,
        version=version,
        timeout=timeout,
        platform=platform,
        config=config,
        cors=cors,
        env=env,
        verbose=verbose,
        public=public,
    )


@serve.group(help='Utility commands for lc-serve.')
@click.help_option('-h', '--help')
def util():
    pass


@util.command(help='Upload a DataFrame to JCloud.')
@click.argument(
    'module',
    type=str,
    required=True,
)
@click.option(
    '--name',
    type=str,
    help='Name of the DataFrame.',
)
def upload_df(module, name):
    upload_df_to_jcloud(module, name)


@util.command(help='Create slack app manifest.')
@click.option(
    '--name',
    type=str,
    default=SLACK_BOT_NAME,
    help='Name of the app.',
    show_default=True,
)
def slack_app_manifest(name):
    from rich.console import Console
    from rich.rule import Rule
    from rich.syntax import Syntax
    from rich.text import Text

    syntax = Syntax(create_slack_app_manifest(name), "yaml")
    console = Console()
    console.print(Rule("App Manifest", style="bold green"))
    console.print(
        Text("Copy this yaml to create your Slack App.", style="bold blue"),
        justify='center',
    )
    console.print(Rule(style="bold green"))
    console.print(syntax)
    console.print(Rule(style="bold green"))


@serve.command(help='List all deployed apps.')
@click.option(
    '--phase',
    type=str,
    default=','.join(
        [
            Phase.Serving,
            Phase.Failed,
            Phase.Starting,
            Phase.Updating,
            Phase.Paused,
        ]
    ),
    help='Deployment phase for the app.',
    show_default=True,
)
@click.option(
    '--name',
    type=str,
    default=None,
    help='Name of the app.',
    show_default=True,
)
@click.help_option('-h', '--help')
@syncify
async def list(phase, name):
    await list_apps_on_jcloud(phase=phase, name=name)


@serve.command(help='Get status of a deployed app.')
@click.argument('app-id')
@click.help_option('-h', '--help')
@syncify
async def status(app_id):
    await get_app_status_on_jcloud(app_id)


@serve.command(help='Remove an app.')
@click.argument('app-id')
@click.help_option('-h', '--help')
@syncify
async def remove(app_id):
    await remove_app_on_jcloud(app_id)


@serve.group(help='Play with predefined apps on JCloud.')
@click.help_option('-h', '--help')
def playground():
    pass


@playground.command(help='Play with babyagi on JCloud.')
@click.option(
    '--verbose',
    is_flag=True,
    help='Verbose mode.',
    show_default=True,
)
def babyagi(verbose):
    sys.path.append(os.path.join(os.path.dirname(__file__), 'playground', 'babyagi'))
    from .playground.babyagi.playground import play

    play(verbose=verbose)


@playground.command(help='Play with autogpt on JCloud.')
@click.option(
    '--verbose',
    is_flag=True,
    help='Verbose mode.',
    show_default=True,
)
@syncify
async def autogpt(verbose):
    sys.path.append(os.path.join(os.path.dirname(__file__), 'playground', 'autogpt'))
    from .playground.autogpt.playground import play

    await play(verbose=verbose)


@playground.command(help='Play with pdf qna on JCloud.')
def pdf_qna():
    try:
        from streamlit.web import cli as strcli
    except ImportError:
        raise ImportError(
            "Streamlit is not installed. Please install it with `pip install streamlit`."
        )

    sys.argv = [
        'streamlit',
        'run',
        os.path.join(
            os.path.dirname(__file__), 'playground', 'pdf_qna', 'playground.py'
        ),
    ]
    sys.exit(strcli.main())


@playground.command(help='Play with pandas-ai on JCloud.')
@click.argument(
    'host',
    type=str,
    required=True,
)
@click.option(
    '--verbose',
    is_flag=True,
    help='Verbose mode.',
    show_default=True,
)
@syncify
async def pandas_ai(host, verbose):
    sys.path.append(os.path.join(os.path.dirname(__file__), 'playground', 'pandas_ai'))
    from .playground.pandas_ai.playground import converse

    await converse(host=host, verbose=verbose)


if __name__ == "__main__":
    serve()



import threading
from contextlib import nullcontext
from typing import Any, Dict, Optional, Union, TYPE_CHECKING

from jina import Executor, requests
from langchain.agents import AgentExecutor, initialize_agent, load_tools
from langchain.chains.loading import load_chain_from_config
from pydantic import BaseModel

if TYPE_CHECKING:
    from docarray import Document, DocumentArray

from .playground.utils.helper import (
    AGENT_OUTPUT,
    CLS,
    DEFAULT_FIELD,
    DEFAULT_KEY,
    LLM_TYPE,
    RESULT,
    Capturing,
    EnvironmentVarCtxtManager,
)


class CombinedMeta(type(Executor), type(BaseModel)):
    def __new__(cls, name, bases, namespace, **kwargs):
        namespace['__fields_set__'] = set()
        return super().__new__(cls, name, bases, namespace, **kwargs)


def _chain_base_model_kwargs(executor_kwargs: Dict, fields: Dict = {}) -> Dict:
    _base_model_kwargs = {}

    def _parse(v):
        # TODO: dynamically load classes
        from langchain.llms import OpenAI
        from langchain.prompts import PromptTemplate

        if isinstance(v, dict):
            cls_str = v.get(CLS)
            if cls_str:
                try:
                    return load_chain_from_config(v['kwargs'])
                except Exception as e:
                    cls = locals()[cls_str]
                    allowed = {field.alias for field in cls.__fields__.values()}
                    _model_kwargs = v.get('kwargs', {})
                    for kk in _model_kwargs.copy().keys():
                        if kk not in allowed:
                            del _model_kwargs[kk]
                    return cls(**_model_kwargs)
            else:
                return {kk: _parse(vv) for kk, vv in v.items()}
        elif isinstance(v, list):
            return [_parse(vv) for vv in v]
        else:
            return v

    try:
        # remove jina specific kwargs by looking at the fields
        _executor_kwargs = {
            k: v for k, v in executor_kwargs.items() if k in fields or k == LLM_TYPE
        }
        _base_model_kwargs = load_chain_from_config(_executor_kwargs)
    except Exception as e:
        for _field, _kwargs in executor_kwargs.items():
            if (fields and _field in fields) or _field in (
                CLS,
                'kwargs',
                LLM_TYPE,
                DEFAULT_FIELD,
            ):
                try:
                    _base_model_kwargs[_field] = load_chain_from_config(_kwargs)
                except Exception as e:
                    _base_model_kwargs[_field] = _parse(_kwargs)

    return _base_model_kwargs


def _agent_base_model_args(executor_kwargs: Dict) -> AgentExecutor:
    def _get_llm_obj(llm_dict: Dict):
        return _chain_base_model_kwargs({'llm': llm_dict}, {'llm': None})['llm']

    def _get_default_llm_obj():
        from langchain.llms import OpenAI

        return OpenAI()

    if 'tools' not in executor_kwargs or not isinstance(executor_kwargs['tools'], dict):
        raise ValueError('tools must be specified in the config')
    else:
        tools = executor_kwargs['tools']
        tools['llm'] = (
            _get_llm_obj(tools['llm']) if tools.get('llm') else _get_default_llm_obj()
        )
        tools = load_tools(**tools)
        executor_kwargs.pop('tools')

    if 'llm' not in executor_kwargs:
        llm = _get_default_llm_obj()
    else:
        llm = _get_llm_obj(executor_kwargs['llm'])
        executor_kwargs.pop('llm')

    if (
        'agent' in executor_kwargs
        and executor_kwargs['agent'] == 'self-ask-with-search'
    ):
        if len(tools) > 1:
            raise ValueError(
                'self-ask-with-search only works with one tool, but got %d' % len(tools)
            )
        if tools[0].name != 'Intermediate Answer':
            tools[0].name = 'Intermediate Answer'

    return initialize_agent(tools=tools, llm=llm, **executor_kwargs)


class ChainExecutor(Executor):
    def __init_parents__(self, chain_cls, *args, **kwargs):
        chain_cls.__init__(
            self, *args, **_chain_base_model_kwargs(kwargs, chain_cls.__fields__)
        )
        super().__init__(*args, **kwargs)

    def _handle_merge(self, docs_map: Dict[str, 'DocumentArray']) -> 'DocumentArray':
        # Merge Documnets with same ID into one Document and create the DocumentArray
        from docarray import DocumentArray

        da = DocumentArray()
        for k, v in docs_map.items():
            for doc in v:
                if doc.id in da:
                    da[doc.id].tags.update(doc.tags)
                else:
                    da.append(doc)
        return da

    @requests(on='/run')
    def __run_endpoint(
        self,
        docs: 'DocumentArray',
        docs_map: Optional[Dict[str, 'DocumentArray']],
        **kwargs,
    ) -> 'DocumentArray':
        if len(docs_map) > 1:
            docs = self._handle_merge(docs_map)

        for doc in docs:
            self.logger.debug(f'calling run on {doc.tags.keys()}')
            if len(self.output_keys) == 1:
                _result = {self.output_keys[0]: self.run(doc.tags)}
            else:
                _result = {RESULT: self(doc.tags)}
            doc.tags.update(_result)
        return docs

    @requests(on='/arun')
    async def __arun_endpoint(
        self,
        docs: 'DocumentArray',
        docs_map: Optional[Dict[str, 'DocumentArray']],
        **kwargs,
    ) -> 'DocumentArray':
        if len(docs_map) > 1:
            docs = self._handle_merge(docs_map)

        for doc in docs:
            self.logger.debug(f'calling run on {doc.tags.keys()}')
            if len(self.output_keys) == 1:
                _result = {self.output_keys[0]: await self.arun(doc.tags)}
            else:
                _result = {RESULT: await self(doc.tags)}
            doc.tags.update(_result)
        return docs


class LangchainAgentExecutor(Executor):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._capture_lock = threading.Lock()

    @staticmethod
    def run_input(doc: 'Document') -> Dict:
        return doc.tags if DEFAULT_KEY not in doc.tags else doc.tags[DEFAULT_KEY]

    def get_capture_ctx(self) -> Capturing:
        return (
            nullcontext()
            if self._capture_lock.locked()
            else Capturing(lock=self._capture_lock)
        )

    def update_agent_output(
        self,
        cap: Union[Capturing, nullcontext],
        doc: 'Document',
        html: bool = False,
    ):
        from ansi2html import Ansi2HTMLConverter

        if isinstance(cap, Capturing):
            if html:
                converter = Ansi2HTMLConverter()
                doc.tags.update({AGENT_OUTPUT: converter.convert(''.join(cap))})
            else:
                doc.tags.update({AGENT_OUTPUT: ''.join(cap)})
        else:
            doc.tags.update({AGENT_OUTPUT: ''})

    @requests(on='/load_and_run')
    def __load_and_run_endpoint(
        self, docs: 'DocumentArray', parameters, **kwargs
    ) -> 'DocumentArray':
        self.logger.info(f'loading agent with {parameters} and docs {docs}')
        with EnvironmentVarCtxtManager(
            parameters['env'] if 'env' in parameters else {}
        ):
            try:
                agent = _agent_base_model_args(parameters)
            except ValueError as e:
                self.logger.error(e)
                for doc in docs:
                    doc.tags.update({RESULT: str(e)})
                return docs

            html = parameters['html'] if 'html' in parameters else False
            for doc in docs:
                self.logger.debug(f'calling run on {doc.tags.keys()}')
                with self.get_capture_ctx() as cap:
                    doc.tags.update({RESULT: agent.run(self.run_input(doc))})
                self.update_agent_output(cap, doc, html)
        return docs

    @requests(on='/aload_and_run')
    async def __aload_and_run_endpoint(
        self, docs: 'DocumentArray', parameters, **kwargs
    ) -> 'DocumentArray':
        self.logger.info(f'loading agent with {parameters} and docs {docs}')
        with EnvironmentVarCtxtManager(
            parameters['env'] if 'env' in parameters else {}
        ):
            try:
                agent = _agent_base_model_args(parameters)
            except ValueError as e:
                self.logger.error(e)
                for doc in docs:
                    doc.tags.update({RESULT: str(e)})
                return docs

            html = parameters['html'] if 'html' in parameters else False
            for doc in docs:
                self.logger.debug(f'calling run on {doc.tags.keys()}')
                with self.get_capture_ctx() as cap:
                    doc.tags.update({RESULT: await agent.arun(self.run_input(doc))})
                self.update_agent_output(cap, doc, html)

        return docs



import json
import os
from functools import lru_cache
from typing import Any, Dict, Generator, List, Tuple, Union
from urllib.parse import urlparse

from jina.logging.logger import JinaLogger
from langchain.agents import ConversationalAgent
from langchain.memory import ChatMessageHistory
from langchain.output_parsers import StructuredOutputParser
from langchain.prompts import PromptTemplate
from langchain.schema import ChatMessage
from langchain.tools import StructuredTool
from langchain.tools.base import ToolException
from slack_sdk import WebClient
from slack_sdk.errors import SlackApiError
from tenacity import retry, stop_after_attempt, wait_exponential

PROGRESS_MESSAGE = "Processing..."


class SlackBot:
    _logger = JinaLogger('SlackBot')

    def __init__(self, workspace: str):
        from langchain.output_parsers import PydanticOutputParser
        from slack_bolt import App
        from slack_bolt.adapter.fastapi import SlackRequestHandler

        try:
            from helper import TextOrBlock
        except ImportError:
            from .helper import TextOrBlock

        self.slack_app = App()
        self.workspace = workspace
        self.handler = SlackRequestHandler(self.slack_app)
        self._parser = PydanticOutputParser(pydantic_object=TextOrBlock)

    @staticmethod
    def slack_client() -> WebClient:
        return WebClient(token=os.environ.get('SLACK_BOT_TOKEN'))

    @staticmethod
    def get_username(userid: str) -> str:
        try:
            response = SlackBot.slack_client().users_profile_get(user=userid)
            return response.data['profile']['real_name']
        except Exception as e:
            return None

    @classmethod
    def extract_channel_ts(cls, url):
        try:
            parsed_url = urlparse(url)
            if not all([parsed_url.scheme, parsed_url.netloc, parsed_url.path]):
                return None, None

            path_parts: List[str] = parsed_url.path.split('/')
            if len(path_parts) != 4:
                return None, None

            channel_id = path_parts[2]
            thread_ts = (
                path_parts[3].replace('p', '', 1)[:10]
                + '.'
                + path_parts[3].replace('p', '', 1)[10:]
            )

            return channel_id, thread_ts

        except Exception as e:
            cls._logger.error(f"Error extracting channel and ts from url: {e}")
            return None, None

    @classmethod
    def get_history(cls, channel: str, ts: str) -> ChatMessageHistory:
        cls._logger.debug(f"Getting history for {channel} {ts}")

        response = cls.slack_client().conversations_replies(channel=channel, ts=ts)
        msgs: List[Dict] = response["messages"]
        history = ChatMessageHistory()

        def _extract_text_from_blocks(user: str, blocks: Union[List, Dict]):
            if isinstance(blocks, dict):
                for key, value in blocks.items():
                    if key == 'text' and isinstance(value, dict):
                        history.add_message(
                            ChatMessage(
                                content=value['text'],
                                role=user,
                                additional_kwargs={"id": user},
                            )
                        )
                    elif key == 'text' and isinstance(value, str):
                        history.add_message(
                            ChatMessage(
                                content=value,
                                role=user,
                                additional_kwargs={"id": user},
                            )
                        )
                    else:
                        _extract_text_from_blocks(user=user, blocks=value)

            elif isinstance(blocks, list):
                for item in blocks:
                    _extract_text_from_blocks(user=user, blocks=item)

        # read all but the last message
        for msg in msgs[:-1]:
            if msg.get("type") != "message":
                # TODO: not sure how to handle this
                continue

            if 'blocks' in msg:
                if 'user' in msg:
                    username = SlackBot.get_username(msg['user']) or msg['user']
                    user = f"Human ({username})"
                elif 'bot_id' in msg:
                    user = msg['bot_id']

                _extract_text_from_blocks(user=user, blocks=msg['blocks'])

            text: str = msg.get("text")
            if 'bot_id' in msg:
                if text.strip() in ("", PROGRESS_MESSAGE):
                    continue

                history.add_message(
                    ChatMessage(
                        content=text, role="AI", additional_kwargs={"id": msg["bot_id"]}
                    )
                )
            elif 'user' in msg:
                username = SlackBot.get_username(msg['user']) or msg['user']
                history.add_message(
                    ChatMessage(
                        content=text,
                        role=f"Human ({username})",
                    )
                )

        return history

    @classmethod
    def slack_messages(cls, url: str) -> str:
        """\
Get chat messages from an existing slack conversation url. \
It is important to note that this URL should already be present in the conversation history, in the format `https://<workspace>.slack.com/archives/<channel_id>/<thread_ts>`. \
You are not permitted to generate or make up these URLs. \
If you can't find the url, please ask the user to provide it to you.
"""

        cls._logger.debug(f"Getting slack messages from {url}")

        if url.startswith('url='):
            url = url[4:]
        # if url is wrapped with '' or "" or <>, remove them
        if url.startswith("'") and url.endswith("'"):
            url = url[1:-1]
        elif url.startswith('"') and url.endswith('"'):
            url = url[1:-1]
        elif url.startswith('<') and url.endswith('>'):
            url = url[1:-1]

        channel, ts = SlackBot.extract_channel_ts(url)
        if channel is None or ts is None:
            raise ToolException(
                f"Invalid URL `{url}` received, could not extract channel and ts"
            )

        try:
            history = SlackBot.get_history(channel, ts)
        except Exception as e:
            _err_msg = (
                f"Invalid URL `{url}` received, could not extract channel and ts as {e}"
            )
            if isinstance(e, SlackApiError):
                if e.response["error"] == "not_in_channel":
                    _err_msg = f"Cannot access the channel `{channel}`. Please add me to the channel and try again."
                elif e.response["error"] == "channel_not_found":
                    _err_msg = f"Channel `{channel}` was not found. Please check the URL and try again."
                elif e.response["error"] == "thread_not_found":
                    _err_msg = f"Thread `{ts}` was not found. Please check the URL and try again."

            raise ToolException(_err_msg)

        return json.dumps([{msg.role: msg.content} for msg in history.messages])

    @staticmethod
    @retry(stop=stop_after_attempt(3), wait=wait_exponential(multiplier=1, max=10))
    def send_message(
        client: WebClient,
        channel: str,
        ts: str,
        text: str = None,
        blocks: List[Dict] = None,
    ) -> Tuple[str, str]:
        if text is not None:
            response = client.chat_postMessage(channel=channel, thread_ts=ts, text=text)
        elif blocks is not None:
            response = client.chat_postMessage(
                channel=channel, thread_ts=ts, blocks=blocks
            )
        else:
            raise ValueError("Either text or blocks must be specified")
        return response["channel"], response["ts"]

    @staticmethod
    @retry(stop=stop_after_attempt(3), wait=wait_exponential(multiplier=1, max=10))
    def update_message(
        client: WebClient,
        channel: str,
        ts: str,
        text: str = None,
        blocks: List[Dict] = None,
    ):
        if text is not None:
            client.chat_update(channel=channel, ts=ts, text=text)
        elif blocks is not None:
            client.chat_update(channel=channel, ts=ts, text=text, blocks=blocks)
        else:
            raise ValueError("Either text or blocks must be specified")

    @staticmethod
    def send(
        client: WebClient,
        channel: str,
        thread_ts: str,
        parser: StructuredOutputParser,
        progress_message: str = PROGRESS_MESSAGE,
    ):
        try:
            from helper import TextOrBlock
        except ImportError:
            from .helper import TextOrBlock

        # send a progress message first on the thread
        channel, ts = SlackBot.send_message(
            client, channel, thread_ts, progress_message
        )

        def __call__(text: Union[str, Generator[str, None, None]]):
            message_text = ""

            if isinstance(text, Generator):
                for i, t in enumerate(text):
                    message_text += t
                    SlackBot.update_message(client, channel, ts, message_text)
            else:
                try:
                    textOrBlock: TextOrBlock = parser.parse(text)
                except Exception as e:
                    SlackBot.update_message(client, channel, ts, text=text)
                    return

                if textOrBlock.kind == "text":
                    SlackBot.update_message(
                        client=client,
                        channel=channel,
                        ts=ts,
                        text=textOrBlock.text,
                    )
                elif textOrBlock.kind == "block":
                    SlackBot.update_message(
                        client=client,
                        channel=channel,
                        ts=ts,
                        text="Answer:",
                        blocks=[b.dict() for b in textOrBlock.blocks],
                    )

        return __call__

    @classmethod
    @lru_cache
    def get_slack_url(cls):
        response = cls.slack_client().auth_test()
        return response["url"]

    @lru_cache(maxsize=128)
    def is_bot_in_channel(self, client: WebClient, channel: str) -> bool:
        try:
            bot_id = client.auth_test()["user_id"]
            response = client.conversations_members(channel=channel)
            return bot_id in response["members"]

        except SlackApiError as e:
            self._logger.error(f"Error while checking if bot is in channel {e}")
            return False

    @staticmethod
    def get_agent_tools() -> List[StructuredTool]:
        return [
            StructuredTool.from_function(
                func=SlackBot.slack_messages,
                handle_tool_error=True,
            )
        ]

    @staticmethod
    def get_agent_prompt() -> PromptTemplate:
        prefix = """
As an AI bot on Slack, your primary objective is to provide substantial assistance to one or more human users within a Slack thread. \
Your mission is to facilitate the completion of tasks through a strategic approach, gathering comprehensive information by posing pertinent questions to refine your understanding of the users' needs. \
Not only should you deliver precise, insightful responses to aid users in task fulfillment, \
but also be proactive in offering innovative solutions and suggestions they may not have considered. \
If a slack url is provided, you can clean it up and pass it to any existing tools. \
If the answer contains `Human (userid)`, replace it with `<@userid>`.

TOOLS:
------

Assistant has access to the following tools:
        """

        suffix = """Begin!

Previous conversation history:
{chat_history}

Human: {input}
{agent_scratchpad}"""

        return ConversationalAgent.create_prompt(
            tools=SlackBot.get_agent_tools(),
            prefix=prefix,
            suffix=suffix,
        )

    def app_mention(self, func):
        @self.slack_app.event('app_mention')
        def wrapper(client: WebClient, body, context):
            _event: Dict = body["event"]
            _channel = _event["channel"]
            _thread_ts = _event.get("thread_ts", _event["ts"])
            _user = _event["user"]
            if "text" in _event:
                _message = _event["text"]
            elif "message" in _event:
                _message = _event["message"]["text"]
                _thread_ts = _event["message"].get("ts", _thread_ts)

            self._logger.info(
                f"App mentioned by user `{_user}` in channel `{_channel}`. Message: `{_message}` "
            )

            if not self.is_bot_in_channel(client, _channel):
                # send a DM to the user to invite the bot to the channel
                client.chat_postMessage(
                    channel=_user,
                    text=f"Unfortunately, I'm not in the channel (ID: {_channel}), you mentioned me in. Please invite me there and try again.",
                )
                return

            func(
                message=_message,
                prompt=SlackBot.get_agent_prompt(),
                history=SlackBot.get_history(_channel, _thread_ts),
                tools=SlackBot.get_agent_tools(),
                reply=SlackBot.send(
                    client=client,
                    channel=_channel,
                    thread_ts=_thread_ts,
                    parser=self._parser,
                ),
                workspace=self.workspace,
                user=_user,
                context=context,
            )

        return wrapper

    def message(self, func):
        @self.slack_app.event('message')
        def wrapper(client, body, context):
            _event: Dict = body["event"]
            _channel = _event["channel"]
            _thread_ts = _event.get("thread_ts", _event["ts"])

            if "text" in _event:
                _message = _event["text"]
            elif "message" in _event:
                _message = _event["message"]["text"]
                _thread_ts = _event["message"].get("ts", _thread_ts)

            self._logger.info(
                f"DM received in channel `{_channel}`. Message: `{_message}` "
            )

            func(
                message=_message,
                prompt=SlackBot.get_agent_prompt(),
                history=SlackBot.get_history(_channel, _thread_ts),
                tools=SlackBot.get_agent_tools(),
                reply=SlackBot.send(
                    client=client,
                    channel=_channel,
                    thread_ts=_thread_ts,
                    parser=self._parser,
                ),
                workspace=self.workspace,
                user=_channel,
                context=context,
            )

        return wrapper

    def register(self, func) -> Any:
        self.app_mention(func)
        self.message(func)
        return func



import asyncio
import inspect
import os
import shutil
import sys
import time
import uuid
from enum import Enum
from functools import cached_property
from importlib import import_module
from importlib.util import module_from_spec, spec_from_file_location
from pathlib import Path
from typing import (
    TYPE_CHECKING,
    Any,
    Callable,
    Dict,
    List,
    Optional,
    Tuple,
    Type,
    Union,
)

from jina import Gateway
from jina.enums import ProtocolType as GatewayProtocolType
from jina.logging.logger import JinaLogger
from jina.serve.runtimes.gateway.composite import CompositeGateway
from jina.serve.runtimes.gateway.http.fastapi import FastAPIBaseGateway
from opentelemetry.trace import get_current_span
from pydantic import BaseModel, Field, ValidationError, create_model
from starlette.types import ASGIApp, Receive, Scope, Send
from websockets.exceptions import ConnectionClosed

from .langchain_helper import (
    AsyncStreamingWebsocketCallbackHandler,
    BuiltinsWrapper,
    OpenAITracingCallbackHandler,
    StreamingWebsocketCallbackHandler,
    TracingCallbackHandler,
)
from .playground.utils.helper import (
    AGENT_OUTPUT,
    APPDIR,
    DEFAULT_KEY,
    LANGCHAIN_API_PORT,
    LANGCHAIN_PLAYGROUND_PORT,
    RESULT,
    SERVING,
    Capturing,
    ChangeDirCtxtManager,
    EnvironmentVarCtxtManager,
    import_from_string,
    parse_uses_with,
    run_cmd,
    run_function,
)

if TYPE_CHECKING:
    from fastapi import FastAPI
    from opentelemetry.sdk.metrics import Counter
    from opentelemetry.trace import Tracer

cur_dir = os.path.dirname(__file__)


class RouteType(str, Enum):
    """RouteType is the type of route"""

    HTTP = 'http'
    WEBSOCKET = 'websocket'


class PlaygroundGateway(Gateway):
    def __init__(self, **kwargs):
        from streamlit.file_util import get_streamlit_file_path

        super().__init__(**kwargs)
        self.streamlit_script = 'playground/app.py'
        # copy playground/config.toml to streamlit config.toml
        streamlit_config_toml_src = os.path.join(cur_dir, 'playground', 'config.toml')
        streamlit_config_toml_dest = get_streamlit_file_path("config.toml")
        # create streamlit_config_toml_dest if it doesn't exist
        os.makedirs(os.path.dirname(streamlit_config_toml_dest), exist_ok=True)
        shutil.copyfile(streamlit_config_toml_src, streamlit_config_toml_dest)

    async def setup_server(self):
        import streamlit.web.bootstrap
        from streamlit.web.server import Server as StreamlitServer

        streamlit.web.bootstrap._fix_sys_path(self.streamlit_script)
        streamlit.web.bootstrap._fix_sys_path(os.path.join(cur_dir, 'playground'))
        streamlit.web.bootstrap._fix_matplotlib_crash()
        streamlit.web.bootstrap._fix_tornado_crash()
        streamlit.web.bootstrap._fix_sys_argv(self.streamlit_script, ())
        streamlit.web.bootstrap._fix_pydeck_mapbox_api_warning()
        streamlit_cmd = f'streamlit run {self.streamlit_script}'
        self.streamlit_server = StreamlitServer(
            os.path.join(cur_dir, self.streamlit_script), streamlit_cmd
        )

    async def run_server(self):
        import streamlit.web.bootstrap

        await self.streamlit_server.start()
        streamlit.web.bootstrap._on_server_start(self.streamlit_server)
        streamlit.web.bootstrap._set_up_signal_handler(self.streamlit_server)

    async def shutdown(self):
        self.streamlit_server.stop()


class LangchainFastAPIGateway(FastAPIBaseGateway):
    @property
    def app(self):
        from docarray import Document, DocumentArray
        from fastapi import Body, FastAPI

        app = FastAPI()

        @app.post("/run")
        async def __run(
            text: str = Body(
                default=...,
                description="The text to be processed",
            ),
            parameters: Dict = Body(
                default=..., description="The parameters to be passed to the executor"
            ),
            envs: Dict = Body(
                default={},
                description="The environment variables to be passed to the executor",
            ),
            executor: str = Body(
                default="agent",
                description="The name of the executor",
            ),
            html: bool = Body(
                default=False,
                description="Whether to return the result as html or plain text",
            ),
        ):
            _parameters = parse_uses_with(parameters)
            if envs and 'env' in _parameters:
                _parameters['env'].update(envs)
            elif envs:
                _parameters['env'] = envs

            _parameters['html'] = html

            da = await self.executor[executor].post(
                on='/load_and_run',
                inputs=DocumentArray([Document(tags={DEFAULT_KEY: text})]),
                parameters=_parameters,
            )
            tags = da[0].tags
            if AGENT_OUTPUT in tags and RESULT in tags:
                return {
                    'result': tags[RESULT],
                    'chain_of_thought': ''.join(tags[AGENT_OUTPUT]),
                }
            elif RESULT in tags:
                return {'result': tags[RESULT]}
            else:
                return {'result': tags}

        @app.get("/healthz")
        async def __healthz():
            return {'status': 'ok'}

        return app


class LangchainAgentGateway(CompositeGateway):
    """The LangchainAgentGateway assumes that the gateway has been started with http on port 8081.
    This is the port on which the nginx process listens. After nginx has been started,
    it will start the playground on port 8501 and the API on port 8080. The actual
    HTTP gateway will start on port 8082.
    Nginx is configured to route the requests in the following way:
    - /playground -> playground on port 8501
    - /api -> API on port 8080
    - / -> HTTP gateway on port 8082
    """

    def __init__(self, **kwargs):
        # need to update port ot 8082, as nginx will listen on 8081
        http_idx = kwargs['runtime_args']['protocol'].index(GatewayProtocolType.HTTP)
        http_port = kwargs['runtime_args']['port'][http_idx]
        if kwargs['runtime_args']['port'][http_idx] != 8081:
            raise ValueError(
                f'Please, let http port ({http_port}) be 8081 for nginx to work'
            )

        kwargs['runtime_args']['port'][http_idx] = 8082
        runtime_args = kwargs['runtime_args']
        runtime_args['metrics_registry'] = None
        runtime_args['tracer_provider'] = None
        runtime_args['grpc_tracing_server_interceptors'] = None
        runtime_args['aio_tracing_client_interceptors'] = None
        runtime_args['tracing_client_interceptor'] = None
        kwargs['runtime_args'] = runtime_args
        super().__init__(**kwargs)

        # remove potential clashing arguments from kwargs
        kwargs.pop("port", None)
        kwargs.pop("protocol", None)

        # note order is important
        self._add_gateway(
            LangchainFastAPIGateway,
            LANGCHAIN_API_PORT,
            **kwargs,
        )
        self._add_gateway(
            PlaygroundGateway,
            LANGCHAIN_PLAYGROUND_PORT,
            **kwargs,
        )

        self.setup_nginx()
        self.nginx_was_shutdown = False

    async def shutdown(self):
        await super().shutdown()
        if not self.nginx_was_shutdown:
            self.shutdown_nginx()
            self.nginx_was_shutdown = True

    def setup_nginx(self):
        command = [
            'nginx',
            '-c',
            os.path.join(cur_dir, '', 'nginx.conf'),
        ]
        output, error = self._run_nginx_command(command)
        self.logger.info('Nginx started')
        self.logger.info(f'nginx output: {output}')
        self.logger.info(f'nginx error: {error}')

    def shutdown_nginx(self):
        command = ['nginx', '-s', 'stop']
        output, error = self._run_nginx_command(command)
        self.logger.info('Nginx stopped')
        self.logger.info(f'nginx output: {output}')
        self.logger.info(f'nginx error: {error}')

    def _run_nginx_command(self, command: List[str]) -> Tuple[bytes, bytes]:
        self.logger.info(f'Running command: {command}')
        output, error = run_cmd(command)
        if error != b'':
            # on CI we need to use sudo; using NOW_CI_RUN isn't good if running test locally
            self.logger.info(f'nginx error: {error}')
            command.insert(0, 'sudo')
            self.logger.info(f'So running command: {command}')
            output, error = run_cmd(command)
        time.sleep(10)
        return output, error

    def _add_gateway(self, gateway_cls, port, protocol='http', **kwargs):
        # ignore metrics_registry since it is not copyable
        runtime_args = self._deepcopy_with_ignore_attrs(
            self.runtime_args,
            [
                'metrics_registry',
                'tracer_provider',
                'grpc_tracing_server_interceptors',
                'aio_tracing_client_interceptors',
                'tracing_client_interceptor',
            ],
        )
        runtime_args.port = [port]
        runtime_args.protocol = [protocol]
        gateway_kwargs = {k: v for k, v in kwargs.items() if k != 'runtime_args'}
        gateway_kwargs['runtime_args'] = dict(vars(runtime_args))
        gateway = gateway_cls(**gateway_kwargs)
        gateway.streamer = self.streamer
        self.gateways.insert(0, gateway)


class ServingGateway(FastAPIBaseGateway):
    def __init__(
        self,
        modules: Tuple[str] = None,
        fastapi_app_str: str = None,
        lcserve_app: bool = False,
        *args,
        **kwargs,
    ):
        super().__init__(*args, **kwargs)
        self._modules = modules
        self._fastapi_app_str = fastapi_app_str
        self._lcserve_app = lcserve_app
        self._fix_sys_path()
        self._init_fastapi_app()
        self._configure_cors()
        self._register_healthz()
        self._register_modules()
        self._setup_metrics()
        self._setup_logging()

    @property
    def app(self) -> 'FastAPI':
        return self._app

    @cached_property
    def workspace(self) -> str:
        import tempfile

        _temp_dir = tempfile.mkdtemp()
        if 'FLOW_ID' not in os.environ:
            self.logger.debug(f'Using temporary workspace directory: {_temp_dir}')
            return _temp_dir

        try:
            flow_id = os.environ['FLOW_ID']
            namespace = flow_id.split('-')[-1]
            return os.path.join('/data', f'jnamespace-{namespace}')
        except Exception as e:
            self.logger.warning(f'Failed to get workspace directory: {e}')
            return _temp_dir

    def _init_fastapi_app(self):
        from fastapi import FastAPI

        if self._fastapi_app_str is not None:
            self.logger.info(f'Loading app from {self._fastapi_app_str}')
            self._app, _ = import_from_string(self._fastapi_app_str)
        else:
            self._app = FastAPI()

    def _configure_cors(self):
        if self.cors:
            self.logger.info('Enabling CORS')
            from fastapi.middleware.cors import CORSMiddleware

            self._app.add_middleware(
                CORSMiddleware,
                allow_origins=['*'],
                allow_credentials=True,
                allow_methods=['*'],
                allow_headers=['*'],
            )

    def _fix_sys_path(self):
        if os.getcwd() not in sys.path:
            sys.path.append(os.getcwd())
        if Path(APPDIR).exists() and APPDIR not in sys.path:
            # This is where the app code is mounted in the container
            sys.path.append(APPDIR)

        if self._lcserve_app:
            # register all predefined apps to sys.path if they exist
            if os.path.exists(os.path.join(APPDIR, 'lcserve', 'apps')):
                for app in os.listdir(os.path.join(APPDIR, 'lcserve', 'apps')):
                    sys.path.append(os.path.join(APPDIR, 'lcserve', 'apps', app))

    def _setup_metrics(self):
        from opentelemetry.instrumentation.fastapi import FastAPIInstrumentor

        if not self.meter_provider:
            self.http_duration_counter = None
            self.ws_duration_counter = None
            return

        FastAPIInstrumentor.instrument_app(
            self._app,
            meter_provider=self.meter_provider,
            tracer_provider=self.tracer_provider,
        )

        self.duration_counter = self.meter.create_counter(
            name="lcserve_request_duration_seconds",
            description="Lc-serve Request duration in seconds",
            unit="s",
        )

        self.request_counter = self.meter.create_counter(
            name="lcserve_request_count",
            description="Lc-serve Request count",
        )

        self.app.add_middleware(
            MetricsMiddleware,
            duration_counter=self.duration_counter,
            request_counter=self.request_counter,
        )

    def _setup_logging(self):
        self.app.add_middleware(LoggingMiddleware, logger=self.logger)

    def _register_healthz(self):
        @self.app.get("/healthz")
        async def __healthz():
            return {'status': 'ok'}

        @self.app.get("/dry_run")
        async def __dry_run():
            return {'status': 'ok'}

    def _update_dry_run_with_ws(self):
        """Update the dry_run endpoint to a websocket endpoint"""
        from fastapi import WebSocket
        from fastapi.routing import APIRoute

        for route in self.app.routes:
            if route.path == '/dry_run' and isinstance(route, APIRoute):
                self.app.routes.remove(route)
                break

        @self.app.websocket("/dry_run")
        async def __dry_run(websocket: WebSocket):
            await websocket.accept()
            await websocket.send_json({'status': 'ok'})
            await websocket.close()

    def _register_modules(self):
        if self._modules is None:
            return

        self.logger.debug(f'Loading modules/files: {",".join(self._modules)}')
        for mod in self._modules:
            # TODO: add support for registering a directory
            if Path(mod).is_file() and mod.endswith('.py'):
                self._register_file(Path(mod))
            else:
                self._register_mod(mod)

    def _register_mod(self, mod: str):
        try:
            app_module = import_module(mod)
            for _, func in inspect.getmembers(app_module, inspect.isfunction):
                self._register_func(func, dirname=os.path.dirname(app_module.__file__))
        except ModuleNotFoundError as e:
            import traceback

            traceback.print_exc()
            self.logger.error(f'Unable to import module: {mod} as {e}')

    def _register_file(self, file: Path):
        try:
            spec = spec_from_file_location(file.stem, file)
            mod = module_from_spec(spec)
            spec.loader.exec_module(mod)
            for _, func in inspect.getmembers(mod, inspect.isfunction):
                self._register_func(func, dirname=os.path.dirname(file))
        except Exception as e:
            self.logger.error(f'Unable to import {file}: {e}')

    def _register_func(self, func: Callable, dirname: str = None):
        def _get_decorator_params(func):
            if hasattr(func, '__serving__'):
                return getattr(func, '__serving__').get('params', {})
            elif hasattr(func, '__ws_serving__'):
                return getattr(func, '__ws_serving__').get('params', {})
            elif hasattr(func, '__slackbot__'):
                return getattr(func, '__slackbot__').get('params', {})
            return {}

        _decorator_params = _get_decorator_params(func)
        if hasattr(func, '__serving__'):
            self._register_http_route(
                func,
                dirname=dirname,
                auth=_decorator_params.get('auth', None),
                openai_tracing=_decorator_params.get('openai_tracing', False),
            )
        elif hasattr(func, '__ws_serving__'):
            self._register_ws_route(
                func,
                dirname=dirname,
                auth=_decorator_params.get('auth', None),
                include_ws_callback_handlers=_decorator_params.get(
                    'include_ws_callback_handlers', False
                ),
                openai_tracing=_decorator_params.get('openai_tracing', False),
            )
        elif hasattr(func, '__slackbot__'):
            self._register_slackbot(
                func,
                dirname=dirname,
                openai_tracing=_decorator_params.get('openai_tracing', False),
            )

    def _register_http_route(
        self,
        func: Callable,
        dirname: str = None,
        auth: Callable = None,
        openai_tracing: bool = False,
        **kwargs,
    ):
        return self._register_route(
            func,
            dirname=dirname,
            auth=auth,
            route_type=RouteType.HTTP,
            openai_tracing=openai_tracing,
            **kwargs,
        )

    def _register_ws_route(
        self,
        func: Callable,
        dirname: str = None,
        auth: Callable = None,
        include_ws_callback_handlers: bool = False,
        openai_tracing: bool = False,
        **kwargs,
    ):
        return self._register_route(
            func,
            dirname=dirname,
            auth=auth,
            route_type=RouteType.WEBSOCKET,
            include_ws_callback_handlers=include_ws_callback_handlers,
            openai_tracing=openai_tracing,
            **kwargs,
        )

    def _register_route(
        self,
        func: Callable,
        dirname: str = None,
        auth: Callable = None,
        route_type: RouteType = RouteType.HTTP,
        include_ws_callback_handlers: bool = False,
        openai_tracing: bool = False,
        **kwargs,
    ):
        _name = func.__name__.title().replace('_', '')

        # check if _name is already registered
        if _name in [route.name for route in self.app.routes]:
            self.logger.debug(f'Route {_name} already registered. Skipping...')
            return

        class Config:
            arbitrary_types_allowed = True

        _input_fields, _file_fields = _get_input_model_fields(func)

        file_params = _get_file_field_params(_file_fields)
        input_model = create_model(
            f'Input{_name}',
            __config__=Config,
            **_input_fields,
            **{'envs': (Dict[str, str], Field(default={}, alias='envs'))},
        )

        output_model = create_model(
            f'Output{_name}',
            __config__=Config,
            **_get_output_model_fields(func),
        )

        if route_type == RouteType.HTTP:
            self.logger.info(f'Registering HTTP route: {func.__name__}')

            create_http_route(
                app=self.app,
                func=func,
                dirname=dirname,
                auth_func=auth,
                file_params=file_params,
                input_model=input_model,
                output_model=output_model,
                openai_tracing=openai_tracing,
                post_kwargs={
                    'path': f'/{func.__name__}',
                    'name': _name,
                    'description': func.__doc__ or '',
                    'tags': [SERVING],
                },
                workspace=self.workspace,
                logger=self.logger,
                tracer=self.tracer,
            )

        elif route_type == RouteType.WEBSOCKET:
            self.logger.info(f'Registering Websocket route: {func.__name__}')
            self._update_dry_run_with_ws()

            create_websocket_route(
                app=self.app,
                func=func,
                dirname=dirname,
                auth=auth,
                input_model=input_model,
                output_model=output_model,
                ws_kwargs={
                    'path': f'/{func.__name__}',
                    'name': _name,
                },
                include_ws_callback_handlers=include_ws_callback_handlers,
                openai_tracing=openai_tracing,
                workspace=self.workspace,
                logger=self.logger,
                tracer=self.tracer,
            )

    def _register_slackbot(
        self,
        func: Callable,
        dirname: str,
        openai_tracing: bool = False,  # TODO: add openai_tracing to slackbot
        **kwargs,
    ):
        from fastapi import Request

        with ChangeDirCtxtManager(dirname):
            from .slackbot import SlackBot

            self.logger.info(f'Registering slackbot: {func.__name__}')
            bot = SlackBot(workspace=self.workspace)

            @self.app.post("/slack/events")
            async def endpoint(req: Request):
                return await bot.handler.handle(req)

            bot.register(func)


def _get_files_data(kwargs: Dict) -> Dict:
    from fastapi import UploadFile
    from starlette.datastructures import UploadFile as StarletteUploadFile

    _files_data = {}
    for k, v in kwargs.items():
        if isinstance(v, (UploadFile, StarletteUploadFile)):
            _files_data[k] = v

    return _files_data


def _get_func_data(
    func: Callable,
    input_data: Union[str, Dict, BaseModel],
    files_data: Dict,
    auth_response: Any = None,
    workspace: str = None,
    to_support_in_kwargs: Dict = {},
) -> Tuple[Dict[str, Any], Dict[str, str]]:
    import json

    if isinstance(input_data, BaseModel):
        _func_data = dict(input_data)
    elif isinstance(input_data, str):
        _func_data = json.loads(input_data)
    else:
        _func_data = input_data

    _envs = _func_data.pop('envs', {})

    if files_data:
        _func_data.update(files_data)

    # Read functions signature and check if `auth_response` is required or kwargs is present
    _func_params_names = list(inspect.signature(func).parameters.keys())
    if 'auth_response' in _func_params_names:
        _func_data['auth_response'] = auth_response
    elif 'kwargs' in _func_params_names:
        _func_data.update({'auth_response': auth_response})

    # Workspace handle
    if 'workspace' in _func_params_names:
        _func_data['workspace'] = workspace
    elif 'kwargs' in _func_params_names:
        _func_data.update({'workspace': workspace})

    # Populate extra kwargs
    if to_support_in_kwargs and 'kwargs' in _func_params_names:
        _func_data.update(to_support_in_kwargs)

    return _func_data, _envs


def _get_updated_signature(
    file_params: List[inspect.Parameter],
    output_model: BaseModel,
    include_token: bool = False,
) -> inspect.Signature:
    _params = [
        *file_params,
        inspect.Parameter(
            name='input_data',
            kind=inspect.Parameter.POSITIONAL_OR_KEYWORD,
            annotation=str,
        ),
    ]

    if include_token:
        _params.append(
            inspect.Parameter(
                name='token',
                kind=inspect.Parameter.POSITIONAL_OR_KEYWORD,
                annotation=str,
            )
        )
    return inspect.Signature(parameters=_params, return_annotation=output_model)


def create_http_route(
    app: 'FastAPI',
    func: Callable,
    dirname: str,
    auth_func: Callable,
    file_params: List,
    input_model: BaseModel,
    output_model: BaseModel,
    openai_tracing: bool,
    post_kwargs: Dict,
    workspace: str,
    logger: JinaLogger,
    tracer: 'Tracer',
):
    from fastapi import Depends, Form, HTTPException, Security, UploadFile, status
    from fastapi.encoders import jsonable_encoder
    from fastapi.security import HTTPAuthorizationCredentials, HTTPBearer

    bearer_scheme = HTTPBearer()

    async def _the_authorizer(
        credentials: HTTPAuthorizationCredentials = Security(bearer_scheme),
    ) -> Any:
        if not credentials:
            raise HTTPException(
                status_code=status.HTTP_401_UNAUTHORIZED, detail="Bearer token required"
            )

        try:
            auth_response = await run_function(auth_func, token=credentials.credentials)
        except Exception as e:
            logger.error(f'Could not verify token: {e}')
            raise HTTPException(
                status_code=status.HTTP_401_UNAUTHORIZED, detail="Invalid bearer token"
            )

        return auth_response

    async def _the_route(
        input_data: input_model,
        files_data: Dict[str, UploadFile] = {},
        auth_response: Any = None,
    ) -> output_model:
        _output, _error = '', ''
        # Tracing handler provided if kwargs is present
        if openai_tracing:
            to_support_in_kwargs = {
                'tracing_handler': OpenAITracingCallbackHandler(
                    tracer=tracer, parent_span=get_current_span()
                )
            }
        else:
            to_support_in_kwargs = {
                'tracing_handler': TracingCallbackHandler(
                    tracer=tracer, parent_span=get_current_span()
                )
            }

        _func_data, _envs = _get_func_data(
            func=func,
            input_data=input_data,
            files_data=files_data,
            auth_response=auth_response,
            workspace=workspace,
            to_support_in_kwargs=to_support_in_kwargs,
        )
        with EnvironmentVarCtxtManager(_envs), ChangeDirCtxtManager(dirname):
            with Capturing() as stdout:
                try:
                    _output = await run_function(func, **_func_data)
                except Exception as e:
                    logger.error(f'Got an exception: {e}')
                    _error = str(e)

            if _error != '':
                print(f'Error: {_error}')
            return output_model(
                result=_output,
                error=_error,
                stdout='\n'.join(stdout),
            )

    def _the_parser(data: str = Form(...)) -> input_model:
        try:
            model = input_model.parse_raw(data)
        except ValidationError as e:
            raise HTTPException(
                status_code=status.HTTP_422_UNPROCESSABLE_ENTITY,
                detail=jsonable_encoder(e.errors()),
            )

        return model

    if auth_func is not None:
        # If an auth function is present, we need to include the authorizer in the route.

        if len(file_params) > 0:
            # If file params are present, we need to use a custom parser to make sure that
            # the input data included in the Form and parsed correctly.

            async def _the_http_route(
                input_data: input_model = Depends(_the_parser),
                auth_response: Any = Depends(_the_authorizer),
                **kwargs,
            ) -> output_model:
                return await _the_route(
                    input_data=input_data,
                    files_data=_get_files_data(kwargs),
                    auth_response=auth_response,
                )

            _the_http_route.__signature__ = _get_updated_signature(
                file_params, output_model, include_token=True
            )

        else:
            # If no file params are present, we include the input args in the Body.

            async def _the_http_route(
                input_data: input_model, auth_response: Any = Depends(_the_authorizer)
            ) -> output_model:
                return await _the_route(
                    input_data=input_data,
                    files_data={},
                    auth_response=auth_response,
                )

    else:
        # If no auth function is present, no need to include the authorizer in the route.

        if len(file_params) > 0:
            # If file params are present, we need to use a custom parser to make sure that
            # the input data included in the Form and parsed correctly.

            async def _the_http_route(
                input_data: input_model = Depends(_the_parser), **kwargs
            ) -> output_model:
                return await _the_route(
                    input_data=input_data,
                    files_data=_get_files_data(kwargs),
                    auth_response=None,
                )

            _the_http_route.__signature__ = _get_updated_signature(
                file_params, output_model, include_token=False
            )

        else:
            # If no file params are present, we include the input args in the Body.

            async def _the_http_route(input_data: input_model) -> output_model:
                return await _the_route(
                    input_data=input_data,
                    files_data={},
                    auth_response=None,
                )

    # Add the route to the app with POST method
    app.post(**post_kwargs)(_the_http_route)


# TODO: add file upload support for websocket routes
def create_websocket_route(
    app: 'FastAPI',
    func: Callable,
    dirname: str,
    auth: Callable,
    input_model: BaseModel,
    output_model: BaseModel,
    include_ws_callback_handlers: bool,
    openai_tracing: bool,
    ws_kwargs: Dict,
    workspace: str,
    logger: JinaLogger,
    tracer: 'Tracer',
):
    from fastapi import (
        Depends,
        Header,
        WebSocket,
        WebSocketDisconnect,
        WebSocketException,
        status,
    )
    from fastapi.security.utils import get_authorization_scheme_param
    from fastapi.websockets import WebSocketState

    async def _the_authorizer(
        authorization: Union[str, None] = Header(None, alias="Authorization"),
    ) -> Any:
        scheme, token = get_authorization_scheme_param(authorization)
        if not (scheme and token):
            raise WebSocketException(
                code=status.WS_1008_POLICY_VIOLATION, reason="Not authenticated"
            )

        if scheme.lower() != "bearer":
            raise WebSocketException(
                code=status.WS_1008_POLICY_VIOLATION, reason="Unauthorized"
            )

        try:
            auth_response = await run_function(auth, token=token)
        except Exception as e:
            logger.error(f'Could not verify token: {e}')
            raise WebSocketException(
                code=status.WS_1008_POLICY_VIOLATION, reason="Invalid bearer token"
            )

        return auth_response

    async def _the_route(websocket: WebSocket, auth_response: Any = None):
        with BuiltinsWrapper(
            loop=asyncio.get_event_loop(),
            websocket=websocket,
            output_model=output_model,
            wrap_print=False,
        ):

            def _get_error_msg(e: Union[WebSocketDisconnect, ConnectionClosed]) -> str:
                return (
                    f'Client {websocket.client} disconnected from `{func.__name__}` with code {e.code}'
                    + (f' and reason {e.reason}' if e.reason else '')
                )

            await websocket.accept()
            logger.info(f'Client {websocket.client} connected to `{func.__name__}`.')
            _ws_recv_lock = asyncio.Lock()
            try:
                while True:
                    # if websocket is closed, break
                    if websocket.client_state not in [
                        WebSocketState.CONNECTED,
                        WebSocketState.CONNECTING,
                    ]:
                        logger.info(
                            f'Client {websocket.client} already disconnected from `{func.__name__}`. Breaking...'
                        )
                        break

                    async with _ws_recv_lock:
                        _data = await websocket.receive_json()

                    try:
                        _input_data = input_model(**_data)
                    except ValidationError as e:
                        logger.error(
                            f'Exception while converting data to input model: {e}'
                        )
                        _ws_serving_error = str(e)
                        _data = output_model(
                            result='',
                            error=_ws_serving_error,
                        )
                        await websocket.send_text(_data.json())
                        continue

                    # Tracing handler provided if kwargs is present
                    if openai_tracing:
                        to_support_in_kwargs = {
                            'tracing_handler': OpenAITracingCallbackHandler(
                                tracer=tracer, parent_span=get_current_span()
                            )
                        }
                    else:
                        to_support_in_kwargs = {
                            'tracing_handler': TracingCallbackHandler(
                                tracer=tracer, parent_span=get_current_span()
                            )
                        }

                    # If the function is a streaming response, we pass the websocket callback handler,
                    # so that stream data can be sent back to the client.
                    if include_ws_callback_handlers:
                        to_support_in_kwargs.update(
                            {
                                'websocket': websocket,
                                'streaming_handler': StreamingWebsocketCallbackHandler(
                                    websocket=websocket,
                                    output_model=output_model,
                                ),
                                'async_streaming_handler': AsyncStreamingWebsocketCallbackHandler(
                                    websocket=websocket,
                                    output_model=output_model,
                                ),
                            }
                        )
                    _returned_data, _ws_serving_error = '', ''
                    # TODO: add support for file upload
                    _func_data, _envs = _get_func_data(
                        func=func,
                        input_data=_input_data,
                        files_data={},
                        auth_response=auth_response,
                        workspace=workspace,
                        to_support_in_kwargs=to_support_in_kwargs,
                    )
                    with EnvironmentVarCtxtManager(_envs), ChangeDirCtxtManager(
                        dirname
                    ):
                        try:
                            _returned_data = await run_function(func, **_func_data)
                            if inspect.isgenerator(_returned_data):
                                # If the function is a generator, we iterate through the generator and send each item back to the client.
                                for _stream in _returned_data:
                                    _data = output_model(
                                        result=_stream,
                                        error=_ws_serving_error,
                                    )
                                    await websocket.send_text(_data.json())

                            else:
                                # If the function is not a generator, we send the result back to the client.
                                _data = output_model(
                                    result=_returned_data,
                                    error=_ws_serving_error,
                                )
                                await websocket.send_text(_data.json())

                            # Once the generator is exhausted/ function call is completed, send a close message
                            logger.info(
                                f'Closing ws connection `{func.__name__}` for client: {websocket.client}'
                            )
                            await websocket.close()
                            break

                        except (WebSocketDisconnect, ConnectionClosed) as e:
                            logger.info(_get_error_msg(e))
                            break

                        except Exception as e:
                            logger.error(f'Got an exception: {e}', exc_info=True)
                            _ws_serving_error = str(e)
                            # For other errors, we send the error back to the client.
                            _data = output_model(
                                result='',
                                error=_ws_serving_error,
                            )
                            await websocket.send_text(_data.json())

                        if _ws_serving_error != '':
                            print(f'Error: {_ws_serving_error}')

            except (WebSocketDisconnect, ConnectionClosed) as e:
                logger.info(_get_error_msg(e))
                return

    if auth is not None:
        logger.info(f'Auth enabled for `{func.__name__}`')

        @app.websocket(**ws_kwargs)
        async def _create_ws_route(
            websocket: WebSocket, auth_response: Any = Depends(_the_authorizer)
        ) -> output_model:
            return await _the_route(websocket=websocket, auth_response=auth_response)

    else:

        @app.websocket(**ws_kwargs)
        async def _create_ws_route(websocket: WebSocket) -> output_model:
            return await _the_route(websocket=websocket, auth_response=None)


def _get_input_model_fields(
    func: Callable,
) -> Tuple[Dict[str, Tuple[Type, Any]], Dict[str, Tuple[Type, Any]]]:
    from fastapi import UploadFile

    _input_model_fields = {}
    _file_fields = {}

    for _name, _param in inspect.signature(func).parameters.items():
        if _param.kind == inspect.Parameter.VAR_KEYWORD:
            continue

        if _param.annotation is inspect.Parameter.empty:
            raise ValueError(
                f'Annotation missing for parameter {_name} in function {func.__name__}. '
                'Please add type annotations to all parameters.'
            )

        if _param.annotation == UploadFile:
            if _param.default is inspect.Parameter.empty:
                _file_fields[_name] = (_param.annotation, ...)
            else:
                _file_fields[_name] = (_param.annotation, _param.default)
        else:
            if _param.default is inspect.Parameter.empty:
                _input_model_fields[_name] = (_param.annotation, ...)
            else:
                _input_model_fields[_name] = (_param.annotation, _param.default)

    return _input_model_fields, _file_fields


def _get_file_field_params(
    fields: Dict[str, Tuple[Type, Any]]
) -> List[inspect.Parameter]:
    _file_field_params = []
    for _name, _field in fields.items():
        _file_field_params.append(
            inspect.Parameter(
                _name,
                inspect.Parameter.POSITIONAL_ONLY,
                annotation=_field[0],
                default=_field[1],
            )
        )
    return _file_field_params


def _get_output_model_fields(func: Callable) -> Dict[str, Tuple[Type, Any]]:
    def _get_result_type():
        if 'return' in func.__annotations__:
            _return = func.__annotations__['return']
            if hasattr(_return, '__next__'):  # if a Generator, return the first type
                return _return.__next__.__annotations__['return']
            elif _return is None:
                return str
            else:
                return _return
        else:
            return str

    _output_model_fields = {
        'result': (_get_result_type(), ...),
        'error': (str, ...),
        'stdout': (str, Field(default='', alias='stdout')),
    }

    return _output_model_fields


class Timer:
    class SharedData:
        def __init__(self, last_reported_time):
            self.last_reported_time = last_reported_time

    def __init__(self, interval: int):
        self.interval = interval

    async def send_duration_periodically(
        self,
        shared_data: SharedData,
        route: str,
        protocol: str,
        counter: Optional['Counter'] = None,
    ):
        while True:
            await asyncio.sleep(self.interval)
            current_time = time.perf_counter()
            if counter:
                counter.add(
                    current_time - shared_data.last_reported_time,
                    {'route': route, 'protocol': protocol},
                )

            shared_data.last_reported_time = current_time


class MetricsMiddleware:
    def __init__(
        self,
        app: ASGIApp,
        duration_counter: Optional['Counter'] = None,
        request_counter: Optional['Counter'] = None,
    ):
        self.app = app
        self.duration_counter = duration_counter
        self.request_counter = request_counter
        # TODO: figure out solution for static assets
        self.skip_routes = [
            '/docs',
            '/redoc',
            '/openapi.json',
            '/healthz',
            '/dry_run',
            '/metrics',
            '/favicon.ico',
        ]

    async def __call__(self, scope: Scope, receive: Receive, send: Send) -> None:
        if scope['path'] not in self.skip_routes:
            timer = Timer(5)
            shared_data = timer.SharedData(last_reported_time=time.perf_counter())
            send_duration_task = asyncio.create_task(
                timer.send_duration_periodically(
                    shared_data, scope['path'], scope['type'], self.duration_counter
                )
            )
            try:
                await self.app(scope, receive, send)
            finally:
                send_duration_task.cancel()
                if self.duration_counter:
                    self.duration_counter.add(
                        time.perf_counter() - shared_data.last_reported_time,
                        {'route': scope['path'], 'protocol': scope['type']},
                    )
                if self.request_counter:
                    self.request_counter.add(
                        1, {'route': scope['path'], 'protocol': scope['type']}
                    )
        else:
            await self.app(scope, receive, send)


class LoggingMiddleware:
    def __init__(self, app: ASGIApp, logger: JinaLogger):
        self.app = app
        self.logger = logger
        self.skip_routes = [
            '/docs',
            '/redoc',
            '/openapi.json',
            '/healthz',
            '/dry_run',
            '/metrics',
            '/favicon.ico',
        ]

    async def __call__(self, scope: Scope, receive: Receive, send: Send) -> None:
        if scope['path'] not in self.skip_routes:
            # Get IP address, use X-Forwarded-For if set else use scope['client'][0]
            ip_address = scope.get('client')[0] if scope.get('client') else None
            if scope.get('headers'):
                for header in scope['headers']:
                    if header[0].decode('latin-1') == 'x-forwarded-for':
                        ip_address = header[1].decode('latin-1').split(",")[0].strip()
                        break

            # Init the request/connection ID
            request_id = str(uuid.uuid4()) if scope["type"] == "http" else None
            connection_id = str(uuid.uuid4()) if scope["type"] == "websocket" else None

            status_code = None
            start_time = time.perf_counter()

            async def custom_send(message: dict) -> None:
                nonlocal status_code

                # TODO: figure out a way to do the same for ws
                if request_id and message.get('type') == 'http.response.start':
                    message.setdefault('headers', []).append(
                        (b'X-API-Request-ID', str(request_id).encode())
                    )
                    status_code = message.get('status')

                await send(message)

            await self.app(scope, receive, custom_send)

            end_time = time.perf_counter()
            duration = round(end_time - start_time, 3)

            if scope["type"] == "http":
                self.logger.info(
                    f"HTTP request: {request_id} - Path: {scope['path']} - Client IP: {ip_address} - Status code: {status_code} - Duration: {duration} s"
                )
            elif scope["type"] == "websocket":
                self.logger.info(
                    f"WebSocket connection: {connection_id} - Path: {scope['path']} - Client IP: {ip_address} - Duration: {duration} s"
                )

        else:
            await self.app(scope, receive, send)



from typing import List, Union

from langchain.prompts.prompt import PromptTemplate
from langchain.chains import AnalyzeDocumentChain
from langchain.document_loaders import PyPDFLoader
from langchain.chains.question_answering import load_qa_chain


def load_pdf_content(urls: Union[List[str], str]) -> str:
    if isinstance(urls, str):
        urls = [urls]

    all_content = []
    for url in urls:
        if not url.endswith('.pdf'):
            raise ValueError('Only PDFs are supported')
        loader = PyPDFLoader(url)
        pages = loader.load_and_split()
        all_content.append('/n'.join([p.page_content for p in pages]))

    return '/n'.join(all_content)


def get_combine_prompt() -> PromptTemplate:
    combine_prompt_template = """Given the following extracted parts of a long document and a question, determine the language of the question & create a final answer in the language the question was asked in. 
    If you don't know the answer, just say that you don't know. Don't try to make up an answer.

    QUESTION: Which state/country's law governs the interpretation of the contract?
    =========
    Content: This Agreement is governed by English law and the parties submit to the exclusive jurisdiction of the English courts in  relation to any dispute (contractual or non-contractual) concerning this Agreement save that either party may apply to any court for an  injunction or other relief to protect its Intellectual Property Rights.

    Content: No Waiver. Failure or delay in exercising any right or remedy under this Agreement shall not constitute a waiver of such (or any other)  right or remedy.\n\n11.7 Severability. The invalidity, illegality or unenforceability of any term (or part of a term) of this Agreement shall not affect the continuation  in force of the remainder of the term (if any) and this Agreement.\n\n11.8 No Agency. Except as expressly stated otherwise, nothing in this Agreement shall create an agency, partnership or joint venture of any  kind between the parties.\n\n11.9 No Third-Party Beneficiaries.

    Content: (b) if Google believes, in good faith, that the Distributor has violated or caused Google to violate any Anti-Bribery Laws (as  defined in Clause 8.5) or that such a violation is reasonably likely to occur,
    =========
    LANGUAGE: English
    FINAL ANSWER: This Agreement is governed by English law.


    QUESTION: Modiji ne kaunse junction kaa naam rename kiya?
    =========
    Content: Friends, in this very July an interesting endeavour has been undertaken, named Azadi Ki Railgadi Aur Railway Station. The objective of this effort is to make people know the role of Indian Railways in the freedom struggle. There are many such railway stations in the country, which are associated with the history of the freedom movement. You too will be surprised to know about these railway stations. Gomoh Junction in Jharkhand is now officially known as Netaji Subhas Chandra Bose Junction Gomoh. Do you know why? Actually at this very station, Netaji Subhash was successful in dodging British officers by boarding the Kalka Mail. All of you must have heard the name of Kakori Railway Station near Lucknow. The names of bravehearts like Ram Prasad Bismil and Ashfaq Ullah Khan are associated with this station. The brave revolutionaries had displayed their might to the British by looting the treasury of the British being carried by train. Whenever you talk to the people of Tamil Nadu, you will get to know about Vanchi Maniyachchi Junction in Thoothukudi district. This station is named after Tamil freedom fighter Vanchinathan ji. This is the same place where Vanchi, 25 years of age then, had punished one British collector for his actions.
    =========
    LANGUAGE: HINDI, ENGLISH
    FINAL_ANSWER: Gomoh junction, jo ki Jarkhand mai hai, abhi Netaji Subash Chandra Bose Junction ke naam se jaana jaega.


    QUESTION: President ne Michael Jackson ke baare mai kya bola?
    =========
    Content: Madam Speaker, Madam Vice President, our First Lady and Second Gentleman. Members of Congress and the Cabinet. Justices of the Supreme Court. My fellow Americans.  \n\nLast year COVID-19 kept us apart. This year we are finally together again. \n\nTonight, we meet as Democrats Republicans and Independents. But most importantly as Americans. \n\nWith a duty to one another to the American people to the Constitution. \n\nAnd with an unwavering resolve that freedom will always triumph over tyranny. \n\nSix days ago, Russia’s Vladimir Putin sought to shake the foundations of the free world thinking he could make it bend to his menacing ways. But he badly miscalculated. \n\nHe thought he could roll into Ukraine and the world would roll over. Instead he met a wall of strength he never imagined. \n\nHe met the Ukrainian people. \n\nFrom President Zelenskyy to every Ukrainian, their fearlessness, their courage, their determination, inspires the world. \n\nGroups of citizens blocking tanks with their bodies. Everyone from students to retirees teachers turned soldiers defending their homeland.

    Content: And we won’t stop. \n\nWe have lost so much to COVID-19. Time with one another. And worst of all, so much loss of life. \n\nLet’s use this moment to reset. Let’s stop looking at COVID-19 as a partisan dividing line and see it for what it is: A God-awful disease.  \n\nLet’s stop seeing each other as enemies, and start seeing each other for who we really are: Fellow Americans.  \n\nWe can’t change how divided we’ve been. But we can change how we move forward—on COVID-19 and other issues we must face together. \n\nI recently visited the New York City Police Department days after the funerals of Officer Wilbert Mora and his partner, Officer Jason Rivera. \n\nThey were responding to a 9-1-1 call when a man shot and killed them with a stolen gun. \n\nOfficer Mora was 27 years old. \n\nOfficer Rivera was 22. \n\nBoth Dominican Americans who’d grown up on the same streets they later chose to patrol as police officers. \n\nI spoke with their families and told them that we are forever in debt for their sacrifice, and we will carry on their mission to restore the trust and safety every community deserves.

    Content: And a proud Ukrainian people, who have known 30 years  of independence, have repeatedly shown that they will not tolerate anyone who tries to take their country backwards.  \n\nTo all Americans, I will be honest with you, as I’ve always promised. A Russian dictator, invading a foreign country, has costs around the world. \n\nAnd I’m taking robust action to make sure the pain of our sanctions  is targeted at Russia’s economy. And I will use every tool at our disposal to protect American businesses and consumers. \n\nTonight, I can announce that the United States has worked with 30 other countries to release 60 Million barrels of oil from reserves around the world.  \n\nAmerica will lead that effort, releasing 30 Million barrels from our own Strategic Petroleum Reserve. And we stand ready to do more if necessary, unified with our allies.  \n\nThese steps will help blunt gas prices here at home. And I know the news about what’s happening can seem alarming. \n\nBut I want you to know that we are going to be okay.
    =========
    LANGUAGE: HINDI, ENGLISH
    FINAL ANSWER: President ne Michael Jackson ko mention nahi kiya.

    QUESTION: {question}
    =========
    {summaries}
    =========
    LANGUAGE:
    FINAL ANSWER:"""

    return PromptTemplate(
        template=combine_prompt_template, input_variables=["summaries", "question"]
    )


def get_qna_chain(llm) -> AnalyzeDocumentChain:
    comnine_prompt = get_combine_prompt()
    qa_chain = load_qa_chain(
        llm, chain_type="map_reduce", combine_prompt=comnine_prompt
    )
    return AnalyzeDocumentChain(combine_docs_chain=qa_chain)



import asyncio
import json
from typing import Dict, List, Literal, Optional, TypeVar

from langchain.chains import llm


def _patch_colored_text(text: str, color: str) -> str:
    return text


llm.get_colored_text = _patch_colored_text


from fastapi import WebSocket
from langchain import LLMChain, OpenAI, PromptTemplate
from langchain.agents import Tool
from langchain.agents.load_tools import get_all_tool_names, load_tools
from langchain.chains.llm import LLMChain
from langchain.chat_models import ChatOpenAI
from langchain.chat_models.base import BaseChatModel
from langchain.experimental import AutoGPT
from langchain.experimental.autonomous_agents.autogpt.output_parser import (
    AutoGPTOutputParser,
    BaseAutoGPTOutputParser,
)
from langchain.experimental.autonomous_agents.autogpt.prompt import AutoGPTPrompt
from langchain.experimental.autonomous_agents.autogpt.prompt_generator import (
    FINISH_NAME,
)
from langchain.schema import AIMessage, Document, HumanMessage, SystemMessage
from langchain.tools.base import BaseTool
from langchain.vectorstores.base import VectorStoreRetriever
from pydantic import BaseModel, Field, ValidationError

T = TypeVar('T', bound='LCServeAutoGPT')


class LCServeHumanInputRun(BaseTool):
    """Tool that adds the capability to ask user for input."""

    websocket: WebSocket = Field(..., exclude=True)

    name = "Human"
    description = (
        "You can ask a human for guidance when you think you "
        "got stuck or you are not sure what to do next. "
        "The input should be a question for the human."
    )

    def _run(self, query: str) -> str:
        """Use the Human input tool."""
        asyncio.run(self._arun(query))

    async def _arun(self, query: str) -> str:
        """Use the Human tool asynchronously."""
        await self.websocket.send_json({'prompt': query})
        return await self.websocket.receive_text()


def get_default_tools():
    from langchain.agents import Tool
    from langchain.tools.file_management.read import ReadFileTool
    from langchain.tools.file_management.write import WriteFileTool
    from langchain.utilities import SerpAPIWrapper

    search = SerpAPIWrapper()
    tools = [
        Tool(
            name="search",
            func=search.run,
            description="useful for when you need to answer questions about current events. You should ask targeted questions",
        ),
        WriteFileTool(),
        ReadFileTool(),
    ]

    return tools


def get_vectorstore():
    from langchain.docstore import InMemoryDocstore
    from langchain.embeddings import OpenAIEmbeddings
    from langchain.vectorstores import FAISS

    # Define your embedding model
    embeddings_model = OpenAIEmbeddings()
    # Initialize the vectorstore as empty
    import faiss

    embedding_size = 1536
    index = faiss.IndexFlatL2(embedding_size)
    return FAISS(embeddings_model.embed_query, index, InMemoryDocstore({}), {})


class LCServeAutoGPT(AutoGPT):
    def __init__(self, websocket: WebSocket, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.websocket = websocket

    @classmethod
    def from_llm_and_tools(
        cls,
        websocket: WebSocket,
        ai_name: str,
        ai_role: str,
        memory: VectorStoreRetriever,
        tools: List[BaseTool],
        llm: BaseChatModel,
        human_in_the_loop: bool = False,
        output_parser: Optional[BaseAutoGPTOutputParser] = None,
    ) -> T:
        prompt = AutoGPTPrompt(
            ai_name=ai_name,
            ai_role=ai_role,
            tools=tools,
            input_variables=["memory", "messages", "goals", "user_input"],
            token_counter=llm.get_num_tokens,
        )
        chain = LLMChain(
            llm=llm,
            prompt=prompt,
            verbose=True,
            callback_manager=llm.callback_manager,
        )
        return cls(
            websocket,
            ai_name,
            memory,
            chain,
            output_parser or AutoGPTOutputParser(),
            tools,
            feedback_tool=LCServeHumanInputRun(websocket=websocket)
            if human_in_the_loop
            else None,
        )

    async def arun(self, goals: List[str]) -> str:
        user_input = (
            "Determine which next command to use, "
            "and respond using the format specified above:"
        )
        # Interaction Loop
        loop_count = 0
        while True:
            # Discontinue if continuous limit is reached
            loop_count += 1

            # Send message to AI, get response
            assistant_reply = await self.chain.arun(
                goals=goals,
                messages=self.full_message_history,
                memory=self.memory,
                user_input=user_input,
            )

            try:
                _assistant_reply = json.loads(assistant_reply)
            except json.decoder.JSONDecodeError:
                _assistant_reply = assistant_reply

            await self.websocket.send_json(_assistant_reply)
            self.full_message_history.append(HumanMessage(content=user_input))
            self.full_message_history.append(AIMessage(content=assistant_reply))

            # Get command name and arguments
            action = self.output_parser.parse(assistant_reply)
            tools = {t.name: t for t in self.tools}
            if action.name == FINISH_NAME:
                return action.args["response"]
            if action.name in tools:
                tool = tools[action.name]
                try:
                    observation = await tool.arun(action.args)
                except NotImplementedError:
                    observation = tool.run(action.args)
                except ValidationError as e:
                    observation = (
                        f"Validation Error in args: {str(e)}, args: {action.args}"
                    )
                except Exception as e:
                    observation = (
                        f"Error: {str(e)}, {type(e).__name__}, args: {action.args}"
                    )
                result = f"Command {tool.name} returned: {observation}"
            elif action.name == "ERROR":
                result = f"Error: {action.args}. "
            else:
                result = (
                    f"Unknown command '{action.name}'. "
                    f"Please refer to the 'COMMANDS' list for available "
                    f"commands and only respond in the specified JSON format."
                )

            memory_to_add = (
                f"Assistant Reply: {assistant_reply} " f"\nResult: {result} "
            )
            if self.feedback_tool is not None:
                feedback = await self.feedback_tool.arun(
                    'Waiting for feedback, OR pass "exit" to exit.'
                )
                if feedback in {"q", "stop", "exit"}:
                    print("EXITING")
                    return "EXITING"
                memory_to_add += feedback

            self.memory.add_documents([Document(page_content=memory_to_add)])
            self.full_message_history.append(SystemMessage(content=result))


class CustomTool(BaseModel):
    name: str
    prompt: str
    description: str

    def to_tool(self, llm: OpenAI) -> Tool:
        return Tool(
            name=self.name,
            func=LLMChain(
                llm=llm, prompt=PromptTemplate.from_template(self.prompt)
            ).run,
            description=self.description,
        )


class PredefinedTools(BaseModel):
    names: List[Literal[tuple(get_all_tool_names())]]
    params: Dict[str, str]


def get_tools(
    llm: OpenAI,
    predefined_tools: PredefinedTools,
    custom_tools: List[CustomTool],
) -> List[Tool]:
    if predefined_tools is None or getattr(predefined_tools, 'names') is None:
        lc_tools = get_default_tools()
    else:
        lc_tools = load_tools(
            tool_names=predefined_tools.names, llm=llm, **predefined_tools.params
        )

    if custom_tools is not None:
        for tool in custom_tools:
            lc_tools.append(tool.to_tool(llm))

    return lc_tools


def get_agent(
    name: str,
    role: str,
    websocket: WebSocket,
    tools: List[Tool],
    llm: ChatOpenAI,
    human_in_the_loop: bool = False,
) -> LCServeAutoGPT:
    vectorstore = get_vectorstore()
    return LCServeAutoGPT.from_llm_and_tools(
        websocket=websocket,
        ai_name=name,
        ai_role=role,
        tools=tools,
        llm=llm,
        memory=vectorstore.as_retriever(),
        human_in_the_loop=human_in_the_loop,
    )



import asyncio
import os
import subprocess
import sys
import threading
import importlib
import uuid
from collections import defaultdict
from io import StringIO
from typing import Any, Dict, List, Union, Callable, Tuple
import inspect
import functools

import nest_asyncio
from pydantic import BaseModel

CLS = 'cls'
RESULT = 'result'
LLM_TYPE = '_type'
DEFAULT_FIELD = 'chain'
DEFAULT_KEY = '__default__'
AGENT_OUTPUT = '__agent_output__'
SERVING = 'Serving'
APPDIR = '/appdir'

LANGCHAIN_API_PORT = os.environ.get('LANGCHAIN_API_PORT', 8080)
LANGCHAIN_PLAYGROUND_PORT = os.environ.get('LANGCHAIN_PLAYGROUND_PORT', 8501)


try:
    nest_asyncio.apply()
except RuntimeError:
    loop = asyncio.new_event_loop()
    asyncio.set_event_loop(loop)

    try:
        nest_asyncio.apply()
    except RuntimeError:
        pass


def get_or_create_eventloop():
    try:
        return asyncio.get_event_loop()
    except RuntimeError as ex:
        if "There is no current event loop in thread" in str(ex):
            loop = asyncio.new_event_loop()
            asyncio.set_event_loop(loop)
            return asyncio.get_event_loop()


def asyncio_run(func, *args, **kwargs):
    return get_or_create_eventloop().run_until_complete(func(*args, **kwargs))


def asyncio_run_property(func):
    task = asyncio.ensure_future(func)
    get_or_create_eventloop().run_until_complete(task)
    return task.result()


def parse_uses_with(uses_with: Union[Dict, BaseModel, List]) -> Dict[str, Any]:
    _uses_with = defaultdict(dict)

    def _parse(v):
        if isinstance(v, BaseModel):
            return {'cls': v.__class__.__name__, 'kwargs': v.dict(exclude_unset=True)}
        elif isinstance(v, type):
            return {'cls': v.__name__}
        elif isinstance(v, dict):
            return (
                {'cls': v['cls'], 'kwargs': v['kwargs']}
                if 'cls' in v
                else {kk: _parse(vv) for kk, vv in v.items()}
            )
        elif isinstance(v, list):
            return [_parse(vv) for vv in v]
        elif isinstance(v, (str, bool)):
            return v
        elif v is None:
            return v
        else:
            return v

    if isinstance(uses_with, BaseModel):
        uses_with = uses_with.dict()

    if isinstance(uses_with, dict):
        for k, v in uses_with.items():
            _uses_with[k] = _parse(v)
    elif isinstance(uses_with, list):
        for v in uses_with:
            _uses_with.update(_parse(v))

    return _uses_with


class Capturing(list):
    def __enter__(self, lock: threading.Lock = None):
        self._lock = lock
        if self._lock:
            self._lock.acquire()
        self._stdout = sys.stdout
        sys.stdout = self._stringio = StringIO()
        return self

    def __exit__(self, *args):
        self.extend(self._stringio.getvalue().splitlines())
        del self._stringio  # free up some memory
        sys.stdout = self._stdout
        if self._lock:
            self._lock.release()


class EnvironmentVarCtxtManager:
    """a class to wrap env vars"""

    def __init__(self, envs: Dict):
        """
        :param envs: a dictionary of environment variables
        """
        self._env_keys_added: Dict = envs
        self._env_keys_old: Dict = {}

    def __enter__(self):
        for key, val in self._env_keys_added.items():
            # Store the old value, if it exists
            if key in os.environ:
                self._env_keys_old[key] = os.environ[key]
            # Update the environment variable with the new value
            os.environ[key] = str(val)

    def __exit__(self, exc_type, exc_val, exc_tb):
        # Restore the old values of updated environment variables
        for key, val in self._env_keys_old.items():
            os.environ[key] = str(val)
        # Remove any newly added environment variables
        for key in self._env_keys_added.keys():
            os.unsetenv(key)


class ChangeDirCtxtManager:
    """a class to wrap change dir"""

    def __init__(self, path: str):
        """
        :param path: a path to change
        """
        self._path = path
        self._old_path = None

    def __enter__(self):
        self._old_path = os.getcwd()
        os.chdir(self._path)

    def __exit__(self, exc_type, exc_val, exc_tb):
        os.chdir(self._old_path)


def run_cmd(command, std_output=False, wait=True):
    if isinstance(command, str):
        command = command.split()
    if not std_output:
        process = subprocess.Popen(
            command, stdout=subprocess.PIPE, stderr=subprocess.PIPE
        )
    else:
        process = subprocess.Popen(command)
    if wait:
        output, error = process.communicate()
        return output, error


def get_random_tag():
    return 't-' + uuid.uuid4().hex[:5]


def get_random_name():
    return 'n-' + uuid.uuid4().hex[:5]


async def run_function(func: Callable, **kwargs):
    if inspect.iscoroutinefunction(func):
        return await func(**kwargs)
    else:
        return await asyncio.get_running_loop().run_in_executor(
            None,
            functools.partial(func, **kwargs),
        )


class ImportFromStringError(Exception):
    pass


def import_from_string(import_str: Any) -> Tuple[Any, Any]:
    if not isinstance(import_str, str):
        return import_str

    module_str, _, attrs_str = import_str.partition(":")
    if not module_str or not attrs_str:
        message = (
            'Import string "{import_str}" must be in format "<module>:<attribute>".'
        )
        raise ImportFromStringError(message.format(import_str=import_str))

    try:
        module = importlib.import_module(module_str)
    except ImportError as exc:
        if exc.name != module_str:
            raise exc from None
        message = 'Could not import module "{module_str}".'
        raise ImportFromStringError(message.format(module_str=module_str))

    app = module
    try:
        for attr_str in attrs_str.split("."):
            app = getattr(app, attr_str)
    except AttributeError:
        message = 'Attribute "{attrs_str}" not found in module "{module_str}".'
        raise ImportFromStringError(
            message.format(attrs_str=attrs_str, module_str=module_str)
        )

    return app, module



import json
import asyncio
from typing import Dict, Union

import click
import aiohttp
from pydantic import BaseModel, ValidationError

from user_input import UserInput, prompt_user


class Response(BaseModel):
    result: str
    error: str
    stdout: str


class HumanPrompt(BaseModel):
    prompt: str


class ThoughtsCommands(BaseModel):
    thoughts: Dict[str, str]
    command: Dict[str, Union[str, Dict[str, str]]]


async def autogpt(user_input: UserInput, verbose: bool = False):
    _bot = click.style('Bot:', fg='green', bold=True)
    _dim_msg = lambda msg: click.style(msg, fg='green', bold=False, dim=True)
    _thought = click.style('Thought:', fg='green', bold=True, blink=True)
    _command = click.style('Command:', fg='green', bold=True, blink=True)
    _you = click.style('You:', fg='blue', bold=True, blink=True)

    def _echo_if_verbose(msg):
        if verbose:
            click.echo(msg)

    def _bot_said(msg: str, cot=False):
        if cot:
            click.echo(_dim_msg(msg), color=True, nl=False)
        else:
            click.echo(f'\n{_bot} {msg}', color=True, nl=True)

    def _bot_thought(msg: Dict):
        _json_msg = json.dumps(msg, indent=4)
        click.echo(f'\n{_thought} {_json_msg}', color=True)

    def _bot_command(msg: Dict):
        _json_msg = json.dumps(msg, indent=4)
        click.echo(f'\n{_command} {_json_msg}', color=True)

    def _you_prompted():
        return click.prompt(f'{_you}', prompt_suffix=' ')

    try:
        async with aiohttp.ClientSession() as session:
            _url = f'{user_input.host}/{user_input.endpoint}'
            async with session.ws_connect(_url) as ws:
                _bot_said(f'Connected to {_url}.')
                _input_json = {
                    "name": user_input.name,
                    "role": user_input.role,
                    "goals": user_input.goals,
                    "human_in_the_loop": user_input.human_in_the_loop,
                    "envs": user_input.envs,
                }
                _echo_if_verbose(f'Sending: {_input_json}')
                await ws.send_json(_input_json)

                async for msg in ws:
                    if msg.type == aiohttp.WSMsgType.TEXT:
                        _data = msg.data
                        _echo_if_verbose(f'Received: {_data}')

                        if _data == 'close cmd':
                            await ws.close()
                            break
                        else:
                            try:
                                response = Response.parse_raw(_data)
                                _bot_said(response.result, cot=True)
                            except ValidationError:
                                try:
                                    prompt = HumanPrompt.parse_raw(msg.data)
                                    _bot_said(prompt.prompt)
                                    feedback = _you_prompted()
                                    _echo_if_verbose(f'Sending: {feedback}')
                                    await ws.send_str(feedback)
                                except ValidationError:
                                    try:
                                        thoughts_commands = ThoughtsCommands.parse_raw(
                                            msg.data
                                        )
                                        _bot_thought(thoughts_commands.thoughts)
                                        _bot_command(thoughts_commands.command)

                                    except ValidationError:
                                        print(f'Unknown: {msg.data}')

                    elif msg.type == aiohttp.WSMsgType.ERROR:
                        print('ws connection closed with exception %s' % ws.exception())
                    else:
                        print(msg)
    except KeyboardInterrupt:
        pass
    except Exception as e:
        click.echo(e, err=True)


async def play(verbose: bool = False):
    try:
        user_input = prompt_user()
    except ValidationError as e:
        print(e)
        return
    except KeyboardInterrupt:
        return

    await autogpt(user_input=user_input, verbose=verbose)


if __name__ == '__main__':
    asyncio.run(play())
